__author__ = "Simon Nilsson"

import os
from tkinter import *

from simba.mixins.pop_up_mixin import PopUpMixin
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon, FolderSelect,
                                        SimbaButton)
from simba.utils.enums import Formats, Keys, Links
from simba.utils.errors import DuplicationError, NotDirectoryError
from simba.video_processors.batch_process_menus import BatchProcessFrame


class BatchPreProcessPopUp(PopUpMixin):
    """
    :example:
    >>> BatchPreProcessPopUp()
    """

    def __init__(self):
        super().__init__(title="BATCH PROCESS VIDEO", size=(600, 400))
        selections_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="SELECTIONS", icon_name=Keys.DOCUMENTATION.value, icon_link=Links.BATCH_PREPROCESS.value,)
        self.input_folder_select = FolderSelect(selections_frm, "INPUT VIDEO DIRECTORY:", title="Select Folder with Input Videos", lblwidth=30,)
        self.output_folder_select = FolderSelect(selections_frm, "OUTPUT VIDEO DIRECTORY:", title="Select Folder for Output videos", lblwidth=30)

        confirm_btn = SimbaButton(parent=selections_frm, txt="CONFIRM", img='tick', txt_clr='blue', font=Formats.FONT_REGULAR.value, cmd=self.run)
        selections_frm.grid(row=0, column=0, sticky=NW)
        self.input_folder_select.grid(row=0, column=0, sticky=NW)
        self.output_folder_select.grid(row=1, column=0, sticky=NW)
        confirm_btn.grid(row=2, column=0, sticky=NW)
        self.main_frm.mainloop()

    def run(self):
        if not os.path.isdir(self.input_folder_select.folder_path):
            raise NotDirectoryError(msg=f"INPUT folder dir ({self.input_folder_select.folder_path}) is not a valid directory.", source=self.__class__.__name__,)
        if not os.path.isdir(self.output_folder_select.folder_path):
            raise NotDirectoryError(msg=f"OUTPUT folder dir ({self.output_folder_select.folder_path}) is not a valid directory.", source=self.__class__.__name__)
        if (self.output_folder_select.folder_path == self.input_folder_select.folder_path):
            raise DuplicationError(msg="The INPUT directory and OUTPUT directory CANNOT be the same folder", source=self.__class__.__name__,)
        else:
            batch_preprocessor = BatchProcessFrame(
                input_dir=self.input_folder_select.folder_path,
                output_dir=self.output_folder_select.folder_path,
            )
            batch_preprocessor.create_main_window()
            batch_preprocessor.create_video_table_headings()
            batch_preprocessor.create_video_rows()
            batch_preprocessor.create_execute_btn()
            batch_preprocessor.main_frm.mainloop()


#BatchPreProcessPopUp()


__author__ = "Simon Nilsson"

import os
from subprocess import PIPE, Popen
from tkinter import *
from tkinter import filedialog
from typing import Optional, Union

import matplotlib

matplotlib.use('agg')


import cv2
import pandas as pd
from PIL import Image, ImageTk
from tabulate import tabulate

import simba
from simba.mixins.config_reader import ConfigReader
from simba.utils.checks import check_file_exist_and_readable, check_int
from simba.utils.enums import Formats, Options, TagNames
from simba.utils.errors import (AdvancedLabellingError, FrameRangeError,
                                NoDataError)
from simba.utils.lookups import (get_labelling_img_kbd_bindings,
                                 get_labelling_video_kbd_bindings)
from simba.utils.printing import log_event, stdout_success
from simba.utils.read_write import (get_all_clf_names, get_fn_ext,
                                    get_video_meta_data, read_config_entry,
                                    read_df)


class AdvancedLabellingInterface(ConfigReader):
    """
    Launch advanced labelling (annotation) interface in SimBA.

    .. note::
       `Advanced annotation tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/advanced_labelling.md>`__.

    .. image:: _static/img/adv_annotator.png
       :width: 800
       :align: center

    :parameter str config_path: path to SimBA project config file in Configparser format
    :parameter str file_path: Path to video that is to be annotated
    :parameter bool continuing: If True, user is continuing the annotations of a video with started but incomplete annotations.

    Examples
    ----------
    >>> select_labelling_video_advanced(config_path='MyProjectConfig', file_path='MyVideoFilePath', continuing=True)
    """

    def __init__(self, config_path: str, file_path: str, continuing: bool):
        ConfigReader.__init__(self, config_path=config_path)
        log_event(logger_name=str(self.__class__.__name__), log_type=TagNames.CLASS_INIT.value, msg=self.create_log_msg_from_init_args(locals=locals()))
        self.padding, self.file_path = 5, file_path
        self.frm_no, self.video_path = 0, file_path
        self.continuing = continuing
        self.play_video_script_path = os.path.join(os.path.dirname(simba.__file__), "labelling", "play_annotation_video.py")
        _, self.video_name, _ = get_fn_ext(filepath=file_path)
        self.features_extracted_file_path = os.path.join(self.features_dir, self.video_name + "." + self.file_type)
        self.targets_inserted_file_path = os.path.join(self.targets_folder, self.video_name + "." + self.file_type)
        self.machine_results_file_path = os.path.join(self.machine_results_dir, self.video_name + "." + self.file_type)
        self.cap = cv2.VideoCapture(self.video_path)
        self.img_kbd_bindings = get_labelling_img_kbd_bindings()
        self.video_kbd_bindings = get_labelling_video_kbd_bindings()
        self.__create_key_presses_lbl()
        self.__create_video_key_presses_lbl()
        self.video_meta_data = get_video_meta_data(video_path=self.video_path)
        self.frame_lst = list(range(0, self.video_meta_data["frame_count"]))
        self.max_frm_no = max(self.frame_lst)
        self.target_lst = get_all_clf_names(config=self.config, target_cnt=self.clf_cnt)
        if len(self.target_lst) == 0:
            raise NoDataError(msg='To annotate behaviors, your SimBA project needs at least one defined classifier. Found 0', source=self.__class__.__name__)

        self.max_frm_size = 1080, 650
        self.main_window = Toplevel()
        if continuing:
            check_file_exist_and_readable(file_path=self.targets_inserted_file_path)
            check_file_exist_and_readable(file_path=self.features_extracted_file_path)
            if self.file_type == Formats.CSV.value:
                self.data_df = (pd.read_csv(self.targets_inserted_file_path).set_index("Unnamed: 0")[self.target_lst].astype(int))
                self.data_df.index.name = None
            if self.file_type == Formats.PARQUET.value:
                self.data_df = pd.read_parquet(self.targets_inserted_file_path)[self.target_lst]
            self.data_df_features = read_df(self.features_extracted_file_path, self.file_type)
            self.data_df = self.data_df_features.join(self.data_df)
            self.main_window.title("SIMBA ANNOTATION INTERFACE (CONTINUING ANNOTATIONS)")
            try:
                self.frm_no = read_config_entry(self.config, "Last annotated frames", self.video_name.lower(), data_type="int")
            except ValueError:
                pass
        else:
            check_file_exist_and_readable(file_path=self.features_extracted_file_path)
            self.data_df = read_df(self.features_extracted_file_path, self.file_type)
            for target in self.target_lst:
                self.data_df[target] = None
        self.main_window.title("SIMBA ANNOTATION INTERFACE (ADVANCED ANNOTATION) - VIDEO {}".format(self.video_name))
        self.data_df_targets = self.data_df[self.target_lst]
        self.folder = Frame(self.main_window)
        self.buttons_frm = Frame(self.main_window, bd=2, width=700, height=300)
        self.current_frm_n = IntVar(self.main_window, value=self.frm_no)
        self.change_frm_box = Entry(self.buttons_frm, width=7, textvariable=self.current_frm_n)
        self.frame_number_lbl = Label(self.buttons_frm, text="Frame number")
        self.forward_btn = Button(self.buttons_frm, text=">", command=lambda: self.advance_frame(new_frm_number=int(self.current_frm_n.get() + 1)))
        self.backward_btn = Button(self.buttons_frm, text="<", command=lambda: self.advance_frame(new_frm_number=int(self.current_frm_n.get() - 1)))
        self.forward_max_btn = Button(self.buttons_frm, text=">>", command=lambda: self.advance_frame(len(self.frame_lst) - 1))
        self.next_absent_forward_btn = Button(self.buttons_frm, text="NEXT ABSENT", fg="red", command=lambda: self.find_last_next_annotation(forwards=True, present=False))
        self.next_present_forward_btn = Button(self.buttons_frm, text="NEXT PRESENT", fg="blue", command=lambda: self.find_last_next_annotation(forwards=True, present=True))
        self.last_absent_forward_btn = Button(self.buttons_frm, text="PREVIOUS ABSENT", fg="red", command=lambda: self.find_last_next_annotation(forwards=False, present=False))
        self.last_present_forward_btn = Button(self.buttons_frm, text="PREVIOUS PRESENT", fg="blue", command=lambda: self.find_last_next_annotation(forwards=False, present=True))

        self.backward_max_btn = Button(self.buttons_frm, text="<<", command=lambda: self.advance_frame(0))
        self.select_frm_btn = Button(self.buttons_frm, text="Jump to selected frame", command=lambda: self.advance_frame(new_frm_number=int(self.change_frm_box.get())))
        self.jump_frame = Frame(self.main_window)
        self.jump = Label(self.jump_frame, text="Jump Size:")
        self.jump_size = Scale(self.jump_frame, from_=0, to=100, orient=HORIZONTAL, length=200)
        self.jump_size.set(0)
        self.jump_back = Button(self.jump_frame, text="<<", command=lambda: self.advance_frame(int(self.change_frm_box.get()) - self.jump_size.get()))
        self.jump_forward = Button(self.jump_frame, text=">>", command=lambda: self.advance_frame(int(self.change_frm_box.get()) + self.jump_size.get()))

        self.folder.grid(row=0, column=1, sticky=N)
        self.buttons_frm.grid(row=1, column=0)
        self.change_frm_box.grid(row=0, column=1)
        self.last_present_forward_btn.grid(row=1, column=0, sticky=W, padx=self.padding)
        self.last_absent_forward_btn.grid(row=1, column=1, sticky=W, padx=self.padding)
        self.backward_max_btn.grid(row=1, column=2, sticky=W, padx=self.padding)
        self.backward_btn.grid(row=1, column=3, sticky=W, padx=self.padding)
        self.change_frm_box.grid(row=1, column=4)
        self.forward_btn.grid(row=1, column=5, sticky=E, padx=self.padding)
        self.forward_max_btn.grid(row=1, column=6, sticky=W, padx=self.padding)
        self.next_absent_forward_btn.grid(row=1, column=7, sticky=W, padx=self.padding)
        self.next_present_forward_btn.grid(row=1, column=8, sticky=W, padx=self.padding)

        self.select_frm_btn.grid(row=2, column=4, sticky=N)
        self.jump_frame.grid(row=2, column=0)
        self.jump.grid(row=0, column=0, sticky=W)
        self.jump_size.grid(row=0, column=1, sticky=W)
        self.jump_back.grid(row=0, column=2, sticky=E)
        self.jump_forward.grid(row=0, column=3, sticky=W)

        self.check_frame = Frame(self.main_window, bd=2, width=300, height=500)
        self.check_frame.grid(row=0, column=1)
        self.check_behavior_lbl = Label(self.check_frame, text="Check Behavior:")
        self.check_behavior_lbl.config(font=("Calibri", 16))
        self.check_behavior_lbl.grid(row=0, column=0, sticky=N)
        self.behavior_absent_lbl = Label(self.check_frame, text="ABSENT")
        self.behavior_present_lbl = Label(self.check_frame, text="PRESENT")
        self.behavior_present_lbl.grid(row=1, column=1, sticky=N)
        self.behavior_absent_lbl.grid(row=1, column=2, sticky=N)

        self.check_present_vars, self.check_present_checkbox = {}, {}
        self.check_absent_vars, self.check_absent_checkbox = {}, {}
        for target_cnt, target in enumerate(self.target_lst):
            self.check_present_vars[target], self.check_absent_vars[target] = (IntVar(), IntVar())
            self.behavior_name_lbl = Label(self.check_frame, text=target)
            self.check_present_checkbox[target] = Checkbutton(self.check_frame, variable=self.check_present_vars[target], command=lambda: self.save_behavior_in_frm(selection="present"))
            self.check_absent_checkbox[target] = Checkbutton(self.check_frame, variable=self.check_absent_vars[target], command=lambda: self.save_behavior_in_frm(selection="absent"))
            self.behavior_name_lbl.grid(row=target_cnt + 2, column=0, sticky=W)
            self.check_present_checkbox[target].grid(row=target_cnt + 2, column=1, sticky=W)
            self.check_absent_checkbox[target].grid(row=target_cnt + 2, column=2, sticky=W)
            if self.data_df_targets[target].iloc[self.current_frm_n.get()] == 1:
                self.check_present_vars[target].set(value=1)
            elif self.data_df_targets[target].iloc[self.current_frm_n.get()] == 0:
                self.check_absent_vars[target].set(value=1)

        self.range_on = IntVar(value=0)
        self.range_frames = Frame(self.main_window)
        self.range_frames.grid(row=1, column=1, sticky=S)
        self.select_range = Checkbutton(self.range_frames, text="Frame range", variable=self.range_on)
        self.select_range.grid(row=0, column=0, sticky=W)
        self.first_frame = Entry(self.range_frames, width=7)
        self.first_frame.grid(row=0, column=1, sticky=E)
        self.to_label = Label(self.range_frames, text=" to ")
        self.to_label.grid(row=0, column=2, sticky=E)
        self.last_frame = Entry(self.range_frames, width=7)
        self.last_frame.grid(row=0, column=3, sticky=E)

        save = Button(self.main_window, text="Save Range", command=lambda: self.save_behavior_in_range())
        save.grid(row=2, column=1, sticky=N)

        self.generate = Button(self.main_window, text="Save Annotations", command=lambda: self.save_results(), fg="blue")
        self.generate.config(font=("Calibri", 16))
        self.generate.grid(row=10, column=1, sticky=N)

        self.video_player_frm = Frame(self.main_window, width=100, height=100)
        self.video_player_frm.grid(row=0, column=2, sticky=N)
        self.play_video_btn = Button(self.video_player_frm, text="Open Video", command=self.play_video)
        self.play_video_btn.grid(sticky=N, pady=10)
        Label(self.video_player_frm, text=self.video_presses_lbl).grid(sticky=W)
        self.update_img_from_video = Button(self.video_player_frm, text="Show current video frame", command=self.update_frame_from_video)
        self.update_img_from_video.grid(sticky=N)
        self.bind_shortcut_keys()
        Label(self.video_player_frm, text=self.key_presses_lbl).grid(sticky=S)
        self.read_frm(frm_number=0)
        self.main_window.mainloop()

    def __create_key_presses_lbl(self):
        self.key_presses_lbl = "\n\n Keyboard shortcuts for frame navigation: "
        for k, v in self.img_kbd_bindings.items():
            self.key_presses_lbl += '\n'
            self.key_presses_lbl += v['label']

    def bind_shortcut_keys(self):
        self.main_window.bind(self.img_kbd_bindings['save']['kbd'], lambda x: self.save_results())
        self.main_window.bind(self.img_kbd_bindings['frame+1_keep_choices']['kbd'], lambda x: self.advance_frame(new_frm_number=int(self.current_frm_n.get() + 1), keep_prior_img_cb_status=True))
        self.main_window.bind(self.img_kbd_bindings['print_annotation_statistic']['kbd'], lambda x: self.__print_annotation_statistics())
        self.main_window.bind(self.img_kbd_bindings['frame+1']['kbd'], lambda x: self.advance_frame(new_frm_number=int(self.current_frm_n.get() + 1)))
        self.main_window.bind(self.img_kbd_bindings['frame-1']['kbd'], lambda x: self.advance_frame(new_frm_number=int(self.current_frm_n.get() - 1)))
        self.main_window.bind(self.img_kbd_bindings['last_frame']['kbd'], lambda x: self.advance_frame(new_frm_number=self.max_frm_no))
        self.main_window.bind(self.img_kbd_bindings['first_frame']['kbd'], lambda x: self.advance_frame(0))

    def __create_video_key_presses_lbl(self):
        self.video_presses_lbl = "\n\n Keyboard shortcuts for video navigation: "
        for k, v in self.video_kbd_bindings.items():
            self.video_presses_lbl += '\n'
            self.video_presses_lbl += v['label']

    def find_last_next_annotation(self, forwards: bool, present: bool):
        if forwards:
            sliced = self.data_df_targets.loc[self.current_frm_n.get() + 1 :,].sum(
                axis=1
            )
            if present:
                frms = list(sliced.index[sliced >= 1])
            else:
                frms = list(sliced.index[sliced == 0])
            if len(frms) > 0:
                frm = frms[0]
            else:
                raise FrameRangeError(
                    "No forwards frames with annotation detected",
                    source=self.__class__.__name__,
                )
        else:
            sliced = self.data_df_targets.loc[0 : self.current_frm_n.get() - 1,].sum(
                axis=1
            )
            if present:
                frms = list(sliced.index[sliced >= 1])
            else:
                frms = list(sliced.index[sliced == 0])
            if len(frms) > 0:
                frm = frms[-1]
            else:
                raise FrameRangeError(
                    "No backwards frames with annotation detected",
                    source=self.__class__.__name__,
                )

        self.advance_frame(new_frm_number=frm, keep_prior_img_cb_status=False)

    def __print_annotation_statistics(self):
        table_view = [
            ["Video name", self.video_name],
            ["Video frames", self.video_meta_data["frame_count"]],
        ]
        for target in self.target_lst:
            present = sum(
                self.data_df_targets[target][self.data_df_targets[target] == 1]
            )
            absent = sum(
                self.data_df_targets[target][self.data_df_targets[target] == 0]
            )
            table_view.append([target + " present labels", present])
            table_view.append([target + " absent labels", absent])
            table_view.append(
                [target + " % present", present / self.video_meta_data["frame_count"]]
            )
            table_view.append(
                [target + " % absent", absent / self.video_meta_data["frame_count"]]
            )
        headers = ["VARIABLE", "VALUE"]
        print(tabulate(table_view, headers, tablefmt="github"))

    def play_video(self):
        p = Popen(
            "python {}".format(self.play_video_script_path),
            stdin=PIPE,
            stdout=PIPE,
            shell=True,
        )
        main_project_dir = os.path.dirname(self.config_path)
        p.stdin.write(bytes(self.video_path, "utf-8"))
        p.stdin.close()
        temp_file = os.path.join(main_project_dir, "subprocess.txt")
        with open(temp_file, "w") as text_file:
            text_file.write(str(p.pid))

    def update_frame_from_video(self):
        f = open(
            os.path.join(os.path.dirname(self.config_path), "labelling_info.txt"), "r+"
        )
        os.fsync(f.fileno())
        vid_frame_no = int(f.readline())
        self.advance_frame(new_frm_number=vid_frame_no)
        f.close()

    def read_frm(self, frm_number: int):
        self.cap.set(1, frm_number)
        _, self.current_frm_npy = self.cap.read()
        self.current_frm_npy = cv2.cvtColor(self.current_frm_npy, cv2.COLOR_RGB2BGR)
        self.current_frm_pil = Image.fromarray(self.current_frm_npy)
        self.current_frm_pil.thumbnail(self.max_frm_size, Image.ANTIALIAS)
        self.current_frm_pil = ImageTk.PhotoImage(
            master=self.main_window, image=self.current_frm_pil
        )
        self.video_frame = Label(self.main_window, image=self.current_frm_pil)
        self.video_frame.image = self.current_frm_pil
        self.video_frame.grid(row=0, column=0)

    def advance_frame(
        self,
        new_frm_number: int,
        save_behavior_in_previous_frm: bool = True,
        keep_prior_img_cb_status: bool = False,
    ):
        self.check_integrity_of_multiple_classifiers()
        if new_frm_number > self.max_frm_no:
            print(
                "FRAME {} CANNOT BE SHOWN - YOU ARE VIEWING THE FINAL FRAME OF THE VIDEO (FRAME NUMBER {})".format(
                    str(new_frm_number), str(self.max_frm_no)
                )
            )
            self.current_frm_n = IntVar(value=self.max_frm_no)
            self.change_frm_box.delete(0, END)
            self.change_frm_box.insert(0, self.current_frm_n.get())
        elif new_frm_number < 0:
            print(
                "FRAME {} CANNOT BE SHOWN - YOU ARE VIEWING THE FIRST FRAME OF THE VIDEO (FRAME NUMBER {})".format(
                    str(new_frm_number), str(self.max_frm_no)
                )
            )
            self.current_frm_n = IntVar(value=0)
            self.change_frm_box.delete(0, END)
            self.change_frm_box.insert(0, self.current_frm_n.get())
        elif (new_frm_number != self.current_frm_n.get()) and (
            not keep_prior_img_cb_status
        ):
            if save_behavior_in_previous_frm:
                self.save_behavior_in_frm()
            self.create_print_statements()
            self.current_frm_n = IntVar(value=new_frm_number)
            for target in self.target_lst:
                new_frame_annotation = self.data_df_targets[target].loc[
                    int(self.current_frm_n.get())
                ]
                if new_frame_annotation == 0:
                    self.check_absent_vars[target].set(value=1)
                    self.check_present_vars[target].set(value=0)
                elif new_frame_annotation == 1:
                    self.check_present_vars[target].set(value=1)
                    self.check_absent_vars[target].set(value=0)
                else:
                    self.check_present_vars[target].set(value=0)
                    self.check_absent_vars[target].set(value=0)
            self.change_frm_box.delete(0, END)
            self.change_frm_box.insert(0, new_frm_number)
            self.read_frm(frm_number=int(new_frm_number))

        elif (new_frm_number != self.current_frm_n.get()) and (
            keep_prior_img_cb_status
        ):
            self.create_print_statements()
            if save_behavior_in_previous_frm:
                self.save_behavior_in_frm()
            self.current_frm_n = IntVar(value=new_frm_number)
            for target in self.target_lst:
                new_frame_annotation = self.data_df_targets[target].loc[
                    int(self.current_frm_n.get() - 1)
                ]
                if new_frame_annotation == 0:
                    self.check_absent_vars[target].set(value=1)
                elif new_frame_annotation == 1:
                    self.check_present_vars[target].set(value=1)
            self.change_frm_box.delete(0, END)
            self.change_frm_box.insert(0, new_frm_number)
            self.read_frm(frm_number=int(new_frm_number))

    def save_behavior_in_frm(self, selection: str = None):
        for target in self.target_lst:
            target_absent_choice, target_present_choice = (
                self.check_absent_vars[target].get(),
                self.check_present_vars[target].get(),
            )
            if (target_present_choice == 1) & (selection == "present"):
                self.data_df_targets[target].loc[int(self.current_frm_n.get())] = 1
                self.check_absent_vars[target].set(value=0)
            elif (target_absent_choice == 1) & (selection == "absent"):
                self.data_df_targets[target].loc[int(self.current_frm_n.get())] = 0
                self.check_present_vars[target].set(value=0)
            elif (target_absent_choice == 0) & (target_present_choice == 0):
                self.data_df_targets[target].loc[int(self.current_frm_n.get())] = None
            elif target_present_choice == 1:
                self.data_df_targets[target].loc[int(self.current_frm_n.get())] = 1
            elif target_absent_choice == 1:
                self.data_df_targets[target].loc[int(self.current_frm_n.get())] = 0

    def save_behavior_in_range(self):
        self.check_integrity_of_multiple_classifiers()
        start_frm, end_frm = int(self.first_frame.get()), int(self.last_frame.get())
        check_int("START FRAME", int(start_frm), max_value=self.max_frm_no, min_value=0)
        check_int("END FRAME", int(end_frm), max_value=self.max_frm_no, min_value=0)
        if not self.range_on.get():
            raise FrameRangeError(
                msg="TO SAVE RANGE OF FRAMES, TICK THE `Frame range` checkbox before clicking `Save Range`",
                source=self.__class__.__name__,
            )
        elif start_frm < 0:
            raise FrameRangeError(
                msg="FRAME RANGE ERROR: START FRAME {} IS LESS THAN ZERO AND CANNOT BE SHOWN".format(
                    str(start_frm)
                ),
                source=self.__class__.__name__,
            )
        elif end_frm > self.max_frm_no:
            raise FrameRangeError(
                msg="FRAME RANGE ERROR: END FRAME {} IS MORE THAN THE MAX VIDEO FRAME ({}) AND CANNOT BE SHOWN".format(
                    str(end_frm), str(self.max_frm_no)
                ),
                source=self.__class__.__name__,
            )
        elif start_frm == end_frm:
            raise FrameRangeError(
                msg="FRAME RANGE ERROR: START FRAME AND END FRAME IS SET TO THE SAME VALUE ({}) AND DOES NOT REPRESENT A RANGE".format(
                    str(end_frm)
                ),
                source=self.__class__.__name__,
            )
        elif start_frm > end_frm:
            raise FrameRangeError(
                msg="FRAME RANGE ERROR: START FRAME ({}) IS LARGER THAB THE END FRAME ({}). PLEASE SPECIFY A RANGE OF FRAMES WHERE THE START FRAME PRECEDE THE END FRAME".format(
                    str(start_frm), str(end_frm)
                ),
                source=self.__class__.__name__,
            )
        else:
            for frm_no in range(int(start_frm), int(end_frm) + 1):
                for target in self.target_lst:
                    target_absent_choice, target_present_choice = (
                        self.check_absent_vars[target].get(),
                        self.check_present_vars[target].get(),
                    )
                    if target_present_choice == 1:
                        self.data_df_targets[target].loc[frm_no] = 1
                    if target_absent_choice == 1:
                        self.data_df_targets[target].loc[frm_no] = 0
                    if (target_absent_choice == 0) & (target_present_choice == 0):
                        self.data_df_targets[target].loc[frm_no] = None
            self.read_frm(frm_number=int(end_frm))
            self.change_frm_box.delete(0, END)
            self.change_frm_box.insert(0, end_frm)
            self.create_print_statements(
                frame_range=True, start_frame=start_frm, end_frame=end_frm
            )

    def save_results(self):
        self.save_df = read_df(self.features_extracted_file_path, self.file_type)
        self.save_df = pd.concat([self.save_df, self.data_df_targets], axis=1)
        self.save_df = self.save_df.dropna(subset=self.target_lst)
        try:
            if self.file_type == Formats.CSV.value:
                self.save_df.to_csv(self.targets_inserted_file_path)
            if self.file_type == Formats.PARQUET.value:
                self.save_df.to_parquet(self.targets_inserted_file_path)
        except Exception as e:
            print(e, "SIMBA ERROR: File for video {} could not be saved.")
            raise FileExistsError
        stdout_success(
            msg=f"SAVED: Annotation file for video {self.video_name} saved within the project_folder/csv/targets_inserted directory.",
            source=self.__class__.__name__,
        )
        if not self.config.has_section("Last annotated frames"):
            self.config.add_section("Last annotated frames")
        self.config.set(
            "Last annotated frames", str(self.video_name), str(self.current_frm_n.get())
        )
        with open(self.config_path, "w") as configfile:
            self.config.write(configfile)

    def check_integrity_of_multiple_classifiers(self):
        none_target_lst, labelled_target_lst = [], []
        for target in self.target_lst:
            target_absent_choice, target_present_choice = (
                self.check_absent_vars[target].get(),
                self.check_present_vars[target].get(),
            )
            if (target_absent_choice == 0) and (target_present_choice == 0):
                none_target_lst.append(target)
            else:
                labelled_target_lst.append(target)
        if (len(none_target_lst) > 0) & (len(labelled_target_lst) > 0):
            raise AdvancedLabellingError(
                frame=str(self.current_frm_n.get()),
                lbl_lst=labelled_target_lst,
                unlabel_lst=none_target_lst,
                source=self.__class__.__name__,
            )

    def create_print_statements(
        self, frame_range: bool = None, start_frame: int = None, end_frame: int = None
    ):
        print("USER FRAME SELECTION(S):")
        if not frame_range:
            for target in self.target_lst:
                target_absent_choice, target_present_choice = (
                    self.check_absent_vars[target].get(),
                    self.check_present_vars[target].get(),
                )
                if (target_absent_choice == 1) & (target_present_choice == 0):
                    print(
                        "{} ABSENT IN FRAME {}".format(target, self.current_frm_n.get())
                    )
                if (target_present_choice == 1) & (target_absent_choice == 0):
                    print(
                        "{} PRESENT IN FRAME {}".format(
                            target, self.current_frm_n.get()
                        )
                    )
                if (target_present_choice == 0) & (target_absent_choice == 0):
                    print(
                        "{} UN-ANNOTATED IN FRAME {}".format(
                            target, self.current_frm_n.get()
                        )
                    )

        if frame_range:
            for target in self.target_lst:
                target_absent_choice, target_present_choice = (
                    self.check_absent_vars[target].get(),
                    self.check_present_vars[target].get(),
                )
                if (target_present_choice == 1) & (target_absent_choice == 0):
                    print(
                        "{} PRESENT IN FRAMES {} to {}".format(
                            target, str(start_frame), str(end_frame)
                        )
                    )
                elif (target_absent_choice == 1) & (target_present_choice == 0):
                    print(
                        "{} ABSENT IN FRAMES {} to {}".format(
                            target, str(start_frame), str(end_frame)
                        )
                    )
                elif (target_absent_choice == 0) & (target_present_choice == 0):
                    print(
                        "{} UN-ANNOTATED IN FRAMES {} to {}".format(
                            target, str(start_frame), str(end_frame)
                        )
                    )


def select_labelling_video_advanced(
    config_path: Union[str, os.PathLike], continuing: Optional[bool] = False
):
    video_file_path = filedialog.askopenfilename(
        filetypes=[("Video files", Options.ALL_VIDEO_FORMAT_STR_OPTIONS.value)]
    )
    check_file_exist_and_readable(video_file_path)
    video_meta = get_video_meta_data(video_file_path)
    _, video_name, _ = get_fn_ext(video_file_path)
    print(f"ANNOTATING VIDEO {video_name} \n VIDEO INFO: {video_meta}")
    _ = AdvancedLabellingInterface(
        config_path=config_path, file_path=video_file_path, continuing=continuing
    )
#
#
# test = select_labelling_video_advanced(config_path=r"C:\troubleshooting\two_black_animals_14bp\project_folder\project_config.ini",
#                                       continuing=False)



__author__ = "Simon Nilsson"

import os
from typing import List, Optional, Tuple, Union

import numpy as np
import pandas as pd

from simba.mixins.config_reader import ConfigReader
from simba.mixins.feature_extraction_mixin import FeatureExtractionMixin
from simba.mixins.feature_extraction_supplement_mixin import \
    FeatureExtractionSupplemental
from simba.roi_tools.roi_utils import get_roi_dict_from_dfs
from simba.utils.checks import (
    check_all_file_names_are_represented_in_video_log,
    check_file_exist_and_readable, check_float, check_if_dir_exists,
    check_that_column_exist, check_valid_boolean, check_valid_lst)
from simba.utils.data import detect_bouts, slice_roi_dict_for_video
from simba.utils.enums import ROI_SETTINGS, Keys
from simba.utils.errors import CountError, ROICoordinatesNotFoundError
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import get_fn_ext, read_data_paths, read_df
from simba.utils.warnings import NoDataFoundWarning

SHAPE_TYPE = "Shape_type"
TOTAL_ROI_TIME = 'TOTAL ROI TIME (S)'
ENTRY_COUNTS = 'ROI ENTRIES (COUNTS)'
FIRST_ROI_ENTRY_TIME = 'FIRST ROI ENTRY TIME (S)'
LAST_ROI_ENTRY_TIME = 'LAST ROI ENTRY TIME (S)'
MEAN_BOUT_TIME = 'MEAN ROI BOUT TIME (S)'
VELOCITY = 'AVERAGE ROI VELOCITY (CM/S)'
MOVEMENT = 'TOTAL ROI MOVEMENT (CM)'
MEASUREMENT = 'MEASUREMENT'
VIDEO_FPS = 'VIDEO FPS'
VIDEO_LENGTH = 'VIDEO LENGTH (S)'
PIX_PER_MM = 'PIXEL TO MILLIMETER CONVERSION FACTOR'
OUTSIDE_ROI = 'OUTSIDE REGIONS OF INTEREST'


class ROIAggregateStatisticsAnalyzer(ConfigReader, FeatureExtractionMixin):
    """
    Analyzes region-of-interest (ROI) data from video tracking experiments.

    This class computes various statistics related to body-part movements inside defined ROIs, including entry counts, total time spent, and bout durations.

    :param config_path (str | os.PathLike): Path to the configuration file.
    :param data_path (str | os.PathLike | List[str], optional): Path(s) to the data files.
    :param threshold (float): Probability threshold for body-part inclusion.
    :param body_parts (List[str], optional): List of body parts to analyze.
    :param detailed_bout_data (bool): Whether to compute detailed bout data.
    :param calculate_distances (bool): Whether to compute distances traveled.
    :param total_time (bool): Whether to calculate total time spent in ROIs.
    :param entry_counts (bool): Whether to count entries into ROIs.
    :param first_entry_time (bool): Whether to record the first entry time.
    :param outside_rois (bool): If checked, SimBA will treat all areas NOT covered by a ROI drawing as a single additional ROI and compute the chosen metrics for this, single, ROI.
    :param last_entry_time (bool): Whether to record the last entry time.
    :param mean_bout_time (bool): Whether to compute mean bout duration.
    :param transpose (bool): Whether to transpose the final results.
    :param detailed_bout_data_save_path (str | os.PathLike, optional): Path to save detailed bout data.
    :param save_path (str | os.PathLike, optional): Path to save summary statistics.

    :example:
    >>> analyzer = ROIAggregateStatisticsAnalyzer(config_path=r"C:\troubleshooting\mitra\project_folder\project_config.ini", body_parts=['Center'], first_entry_time=True, threshold=0.0, calculate_distances=True, transpose=False, detailed_bout_data=True)
    >>> analyzer.run()
    >>> analyzer.save()
    """


    def __init__(self,
                 config_path: Union[str, os.PathLike],
                 data_path: Optional[Union[str, os.PathLike, List[str]]] = None,
                 threshold: float = 0.0,
                 body_parts: Optional[List[str]] = None,
                 detailed_bout_data: bool = False,
                 calculate_distances: bool = False,
                 total_time: bool = True,
                 entry_counts: bool = True,
                 first_entry_time: bool = False,
                 last_entry_time: bool = False,
                 mean_bout_time: bool = False,
                 outside_rois: bool = False,
                 transpose: bool = False,
                 include_fps: bool = False,
                 include_video_length: bool = False,
                 include_px_per_mm: bool = False,
                 verbose: bool = True,
                 detailed_bout_data_save_path: Optional[Union[str, os.PathLike]] = None,
                 save_path: Optional[Union[str, os.PathLike]] = None):

        check_file_exist_and_readable(file_path=config_path)
        ConfigReader.__init__(self, config_path=config_path)
        if not os.path.isfile(self.roi_coordinates_path):
            raise ROICoordinatesNotFoundError(expected_file_path=self.roi_coordinates_path)
        check_valid_lst(data=body_parts, source=f"{self.__class__.__name__} body-parts", valid_dtypes=(str,), valid_values=self.project_bps)
        check_float(name="Body-part probability threshold", value=threshold, min_value=0.0, max_value=1.0)
        if len(set(body_parts)) != len(body_parts):
            raise CountError(msg=f"All body-part entries have to be unique. Got {body_parts}", source=self.__class__.__name__)
        if detailed_bout_data_save_path is not None:
            check_if_dir_exists(in_dir=os.path.dirname(detailed_bout_data_save_path))
        else:
            detailed_bout_data_save_path = os.path.join(self.logs_path, f'{"Detailed_ROI_data"}_{self.datetime}.csv')
        if save_path is not None:
            check_if_dir_exists(in_dir=os.path.dirname(save_path))
        else:
            save_path = os.path.join(self.logs_path, f'{"ROI_descriptive_statistics"}_{self.datetime}.csv')
        self.detailed_bout_data_save_path, self.save_path = detailed_bout_data_save_path, save_path
        check_valid_boolean(value=[detailed_bout_data], source=f'{self.__class__.__name__} detailed_bout_data', raise_error=True)
        check_valid_boolean(value=[total_time], source=f'{self.__class__.__name__} total_time', raise_error=True)
        check_valid_boolean(value=[entry_counts], source=f'{self.__class__.__name__} entry_counts', raise_error=True)
        check_valid_boolean(value=[first_entry_time], source=f'{self.__class__.__name__} first_entry_time', raise_error=True)
        check_valid_boolean(value=[last_entry_time], source=f'{self.__class__.__name__} last_entry_time', raise_error=True)
        check_valid_boolean(value=[mean_bout_time], source=f'{self.__class__.__name__} mean_bout_time', raise_error=True)
        check_valid_boolean(value=[calculate_distances], source=f'{self.__class__.__name__} calculate_distances', raise_error=True)
        check_valid_boolean(value=[transpose], source=f'{self.__class__.__name__} transpose', raise_error=True)
        check_valid_boolean(value=[include_fps], source=f'{self.__class__.__name__} include_fps', raise_error=True)
        check_valid_boolean(value=[include_video_length], source=f'{self.__class__.__name__} include_video_length', raise_error=True)
        check_valid_boolean(value=[include_px_per_mm], source=f'{self.__class__.__name__} include_px_per_mm', raise_error=True)
        check_valid_boolean(value=[outside_rois], source=f'{self.__class__.__name__} outside_rois', raise_error=True)
        check_valid_boolean(value=[verbose], source=f'{self.__class__.__name__} verbose', raise_error=True)
        self.read_roi_data()
        FeatureExtractionMixin.__init__(self)
        self.data_paths = read_data_paths(path=data_path, default=self.outlier_corrected_paths, default_name=self.outlier_corrected_dir, file_type=self.file_type)
        self.bp_dict, self.bp_lk = {}, {}
        for bp in body_parts:
            animal = self.find_animal_name_from_body_part_name(bp_name=bp, bp_dict=self.animal_bp_dict)
            self.bp_dict[animal] = [f'{bp}_{"x"}', f'{bp}_{"y"}', f'{bp}_{"p"}']
            self.bp_lk[animal] = bp
        self.roi_headers = [v for k, v in self.bp_dict.items()]
        self.roi_headers = [item for sublist in self.roi_headers for item in sublist]
        self.calculate_distances, self.threshold = calculate_distances, threshold
        self.detailed_bout_data = detailed_bout_data
        self.total_time, self.entry_counts = total_time, entry_counts
        self.first_entry_time, self.last_entry_time, self.mean_bout_time, self.include_px_per_mm = first_entry_time, last_entry_time, mean_bout_time, include_px_per_mm
        self.transpose, self.include_fps, self.include_video_length, self.non_roi_zone = transpose, include_fps, include_video_length, outside_rois
        self.detailed_dfs, self.detailed_df, self.verbose = [], [], verbose
        self.results = pd.DataFrame(columns=["VIDEO", "ANIMAL", "BODY-PART", "SHAPE", "SHAPE TYPE", "MEASUREMENT", "VALUE"])

    def __clean_results(self):
        if not self.total_time:
            self.results = self.results[self.results[MEASUREMENT] != TOTAL_ROI_TIME]
        if not self.entry_counts:
            self.results = self.results[self.results[MEASUREMENT] != ENTRY_COUNTS]
        if not self.first_entry_time:
            self.results = self.results[self.results[MEASUREMENT] != FIRST_ROI_ENTRY_TIME]
        if not self.last_entry_time:
            self.results = self.results[self.results[MEASUREMENT] != LAST_ROI_ENTRY_TIME]
        if not self.mean_bout_time:
            self.results = self.results[self.results[MEASUREMENT] != MEAN_BOUT_TIME]
        if not self.calculate_distances:
            self.results = self.results[self.results[MEASUREMENT] != VELOCITY]
            self.results = self.results[self.results[MEASUREMENT] != MOVEMENT]
        if not self.include_fps:
            self.results = self.results[self.results[MEASUREMENT] != VIDEO_FPS]
        if not self.include_video_length:
            self.results = self.results[self.results[MEASUREMENT] != VIDEO_LENGTH]
        if not self.include_px_per_mm:
            self.results = self.results[self.results[MEASUREMENT] != PIX_PER_MM]
        self.results = self.results.sort_values(by=["VIDEO", "ANIMAL", "BODY-PART", "SHAPE", "SHAPE TYPE", "MEASUREMENT"]).reset_index(drop=True)
        if self.transpose:
            self.results['VALUE'] = pd.to_numeric(self.results['VALUE'], errors='coerce')
            self.results = self.results.pivot_table(index=["VIDEO"], columns=["ANIMAL", "SHAPE", "MEASUREMENT"], values="VALUE")
            self.results = self.results.fillna(value='None')
        else:
            self.results = self.results.set_index('VIDEO')

        if self.detailed_bout_data and (len(self.detailed_dfs) > 0):
            self.detailed_df = pd.concat(self.detailed_dfs, axis=0)
            self.detailed_df = self.detailed_df.rename(columns={"Event": "SHAPE NAME", "Start_time": "START TIME", "End Time": "END TIME", "Start_frame": "START FRAME", "End_frame": "END FRAME", "Bout_time": "DURATION (S)"})
            self.detailed_df["BODY-PART"] = self.detailed_df["ANIMAL"].map(self.bp_lk)
            self.detailed_df = self.detailed_df[["VIDEO", "ANIMAL", "BODY-PART", "SHAPE NAME", "START TIME", "END TIME", "START FRAME", "END FRAME", "DURATION (S)"]].reset_index(drop=True)

    def run(self):
        check_all_file_names_are_represented_in_video_log(video_info_df=self.video_info_df, data_paths=self.data_paths)
        for file_cnt, file_path in enumerate(self.data_paths):
            _, video_name, _ = get_fn_ext(file_path)
            video_timer = SimbaTimer(start=True)
            if self.verbose: print(f"Analysing ROI data for video {video_name}... (Video {file_cnt+1}/{len(self.data_paths)})")
            video_settings, pix_per_mm, self.fps = self.read_video_info(video_name=video_name)
            self.sliced_roi_dict, video_shape_names = slice_roi_dict_for_video(data=self.roi_dict, video_name=video_name)
            if len(video_shape_names) == 0:
                NoDataFoundWarning(msg=f"Skipping video {video_name}: No user-defined ROI data found for this video...")
                continue
            else:
                self.sliced_roi_dict = get_roi_dict_from_dfs(rectangle_df=self.sliced_roi_dict[Keys.ROI_RECTANGLES.value],circle_df=self.sliced_roi_dict[Keys.ROI_CIRCLES.value],polygon_df=self.sliced_roi_dict[Keys.ROI_POLYGONS.value])
                self.data_df = read_df(file_path, self.file_type).reset_index(drop=True)
                check_that_column_exist(df=self.data_df, column_name=self.roi_headers, file_name=file_path)
                for animal_name, bp_cols in self.bp_dict.items():
                    p_arr = self.data_df[bp_cols[2]].values.astype(np.float32)
                    below_threshold_idx = np.argwhere(p_arr < self.threshold)
                    bp_arr = self.data_df[[bp_cols[0], bp_cols[1]]].values.astype(np.float32)
                    animal_roi_bouts = []
                    for roi_cnt, (roi_name, roi_data) in enumerate(self.sliced_roi_dict.items()):
                        if roi_data[SHAPE_TYPE].lower() == ROI_SETTINGS.RECTANGLE.value.lower():
                            roi_coords = np.array([[roi_data['topLeftX'], roi_data['topLeftY']], [roi_data['Bottom_right_X'], roi_data['Bottom_right_Y']]])
                            r = FeatureExtractionMixin.framewise_inside_rectangle_roi(bp_location=bp_arr, roi_coords=roi_coords)
                        elif roi_data[SHAPE_TYPE].lower() == ROI_SETTINGS.CIRCLE.value.lower():
                            circle_center = np.array([roi_data['Center_X'], roi_data['Center_Y']]).astype(np.int32)
                            r = FeatureExtractionMixin.is_inside_circle(bp=bp_arr, roi_center=circle_center, roi_radius=roi_data['radius'])
                        else:
                            vertices = roi_data['vertices'].astype(np.int32)
                            r = FeatureExtractionMixin.framewise_inside_polygon_roi(bp_location=bp_arr, roi_coords=vertices)
                        r[below_threshold_idx] = 0
                        self.data_df[roi_data['Name']] = r
                        roi_bouts = detect_bouts(data_df=self.data_df, target_lst=[roi_data["Name"]], fps=self.fps)
                        if len(roi_bouts) > 0:
                            total_time = roi_bouts['Bout_time'].sum()
                            entry_counts = len(roi_bouts)
                            first_entry_time = roi_bouts['Start_time'].values[0]
                            last_entry_time = roi_bouts['Start_time'].values[-1]
                            mean_bout_time = roi_bouts['Bout_time'].mean()
                            movement, velocity = FeatureExtractionSupplemental.movement_stats_from_bouts_df(bp_data=bp_arr, event_name=roi_data["Name"], bout_df=roi_bouts, fps=self.fps, px_per_mm=pix_per_mm)
                            roi_bouts['VIDEO'] = video_name
                            roi_bouts['ANIMAL'] = animal_name
                            self.detailed_dfs.append(roi_bouts)
                            animal_roi_bouts.append(roi_bouts)
                        else:
                            total_time = 0
                            entry_counts = 0
                            first_entry_time = None
                            last_entry_time = None
                            mean_bout_time = None
                            movement, velocity = 0, None
                        self.results.loc[len(self.results)] = [video_name, animal_name, bp_cols[0][:-2], roi_name, roi_data[SHAPE_TYPE], TOTAL_ROI_TIME, total_time]
                        self.results.loc[len(self.results)] = [video_name, animal_name, bp_cols[0][:-2], roi_name, roi_data[SHAPE_TYPE], ENTRY_COUNTS, entry_counts]
                        self.results.loc[len(self.results)] = [video_name, animal_name, bp_cols[0][:-2], roi_name, roi_data[SHAPE_TYPE], FIRST_ROI_ENTRY_TIME, first_entry_time]
                        self.results.loc[len(self.results)] = [video_name, animal_name, bp_cols[0][:-2], roi_name, roi_data[SHAPE_TYPE], LAST_ROI_ENTRY_TIME, last_entry_time]
                        self.results.loc[len(self.results)] = [video_name, animal_name, bp_cols[0][:-2], roi_name, roi_data[SHAPE_TYPE], MEAN_BOUT_TIME, mean_bout_time]
                        self.results.loc[len(self.results)] = [video_name, animal_name, bp_cols[0][:-2], roi_name, roi_data[SHAPE_TYPE], VELOCITY, velocity]
                        self.results.loc[len(self.results)] = [video_name, animal_name, bp_cols[0][:-2], roi_name, roi_data[SHAPE_TYPE], MOVEMENT, movement]
                        self.results.loc[len(self.results)] = [video_name, animal_name, bp_cols[0][:-2], roi_name, roi_data[SHAPE_TYPE], VIDEO_FPS, self.fps]
                        self.results.loc[len(self.results)] = [video_name, animal_name, bp_cols[0][:-2], roi_name, roi_data[SHAPE_TYPE], VIDEO_LENGTH, len(self.data_df) / self.fps]
                        self.results.loc[len(self.results)] = [video_name, animal_name, bp_cols[0][:-2], roi_name, roi_data[SHAPE_TYPE], PIX_PER_MM, pix_per_mm]

                    if self.non_roi_zone and len(animal_roi_bouts) > 0:
                        animal_roi_bouts = pd.concat(animal_roi_bouts, axis=0)
                        animal_roi_bouts = animal_roi_bouts.rename(columns={"Event": "SHAPE NAME", "Start_time": "START TIME", "End Time": "END TIME", "Start_frame": "START FRAME", "End_frame": "END FRAME", "Bout_time": "DURATION (S)"})
                        inside_rois_frm = [i for s, e in zip(animal_roi_bouts['START FRAME'], animal_roi_bouts['END FRAME']) for i in range(s, e+1)]
                        self.data_df[OUTSIDE_ROI] = 1
                        self.data_df.loc[inside_rois_frm, OUTSIDE_ROI] = 0
                        outside_roi_bouts = detect_bouts(data_df=self.data_df, target_lst=OUTSIDE_ROI, fps=self.fps)
                        if len(outside_roi_bouts) > 0:
                            total_time = outside_roi_bouts['Bout_time'].sum()
                            entry_counts = len(outside_roi_bouts)
                            first_entry_time = outside_roi_bouts['Start_time'].values[0]
                            last_entry_time = outside_roi_bouts['Start_time'].values[-1]
                            mean_bout_time = outside_roi_bouts['Bout_time'].mean()
                            movement, velocity = FeatureExtractionSupplemental.movement_stats_from_bouts_df(bp_data=bp_arr, event_name=OUTSIDE_ROI, bout_df=outside_roi_bouts, fps=self.fps, px_per_mm=pix_per_mm)
                            outside_roi_bouts['VIDEO'] = video_name
                            outside_roi_bouts['ANIMAL'] = animal_name
                            self.detailed_dfs.append(outside_roi_bouts)
                            self.results.loc[len(self.results)] = [video_name, animal_name, bp_cols[0][:-2], OUTSIDE_ROI, 'NONE', TOTAL_ROI_TIME, total_time]
                            self.results.loc[len(self.results)] = [video_name, animal_name, bp_cols[0][:-2], OUTSIDE_ROI, 'NONE', ENTRY_COUNTS, entry_counts]
                            self.results.loc[len(self.results)] = [video_name, animal_name, bp_cols[0][:-2], OUTSIDE_ROI, 'NONE', FIRST_ROI_ENTRY_TIME, first_entry_time]
                            self.results.loc[len(self.results)] = [video_name, animal_name, bp_cols[0][:-2], OUTSIDE_ROI, 'NONE', LAST_ROI_ENTRY_TIME, last_entry_time]
                            self.results.loc[len(self.results)] = [video_name, animal_name, bp_cols[0][:-2], OUTSIDE_ROI, 'NONE', MEAN_BOUT_TIME, mean_bout_time]
                            self.results.loc[len(self.results)] = [video_name, animal_name, bp_cols[0][:-2], OUTSIDE_ROI, 'NONE', VELOCITY, velocity]
                            self.results.loc[len(self.results)] = [video_name, animal_name, bp_cols[0][:-2], OUTSIDE_ROI, 'NONE', MOVEMENT, movement]
                            self.results.loc[len(self.results)] = [video_name, animal_name, bp_cols[0][:-2], OUTSIDE_ROI, 'NONE', VIDEO_FPS, self.fps]
                            self.results.loc[len(self.results)] = [video_name, animal_name, bp_cols[0][:-2], OUTSIDE_ROI, 'NONE', VIDEO_LENGTH, len(self.data_df) / self.fps]
                            self.results.loc[len(self.results)] = [video_name, animal_name, bp_cols[0][:-2], OUTSIDE_ROI, 'NONE', PIX_PER_MM, pix_per_mm]

                video_timer.stop_timer()
                if self.verbose: print(f'ROI analysis video {video_name} complete... (elapsed time: {video_timer.elapsed_time_str}s)')


    def save(self):
        self.__clean_results()
        if self.detailed_bout_data and len(self.detailed_df) > 0:
            self.detailed_df.to_csv(self.detailed_bout_data_save_path)
            if self.verbose: print(f"Detailed ROI data saved at {self.detailed_bout_data_save_path}...")
        self.results.to_csv(self.save_path)
        self.timer.stop_timer()
        stdout_success(f'ROI statistics saved at {self.save_path}', elapsed_time=self.timer.elapsed_time_str)


# analyzer = ROIAggregateStatisticsAnalyzer(config_path=r"C:\troubleshooting\mitra\project_folder\project_config.ini",
#                                           body_parts=['Center'],
#                                           include_fps=False,
#                                           threshold=0.5,
#                                           calculate_distances=True,
#                                           transpose=False,
#                                           detailed_bout_data=True,
#                                           outside_rois=True,
#                                           verbose=False)
# analyzer.run()
# analyzer.save()


__author__ = "Simon Nilsson"
__email__ = "sronilsson@gmail.com"

import math
from typing import Optional, Tuple

try:
    from typing import Literal
except:
    from typing_extensions import Literal

import numpy as np
from numba import cuda, int32

try:
    import cupy as cp
except:
    import numpy as cp

from simba.utils.checks import check_float, check_int, check_valid_array
from simba.utils.enums import Formats

THREADS_PER_BLOCK = 1024

@cuda.jit()
def _cuda_direction_from_two_bps(x, y, results):
    i = cuda.grid(1)
    if i > x.shape[0]:
        return
    else:
        a = math.atan2(x[i][0] - y[i][0], y[i][1] - x[i][1]) * (180 / math.pi)
        a = int32(a + 360 if a < 0 else a)
        results[i] = a


def direction_from_two_bps(x: np.ndarray, y: np.ndarray) -> np.ndarray:
    """
    Compute the directionality in degrees from two body-parts. E.g., ``nape`` and ``nose``,
    or ``swim_bladder`` and ``tail`` with GPU acceleration.

    .. csv-table::
       :header: EXPECTED RUNTIMES
       :file: ../../../docs/tables/direction_two_bps.csv
       :widths: 10, 90
       :align: center
       :header-rows: 1

    .. seealso::
       For CPU function see :func:`~simba.mixins.circular_statistics.CircularStatisticsMixin.direction_two_bps`.

    :parameter np.ndarray x: Size len(frames) x 2 representing x and y coordinates for first body-part.
    :parameter np.ndarray y: Size len(frames) x 2 representing x and y coordinates for second body-part.
    :return: Frame-wise directionality in degrees.
    :rtype: np.ndarray.

    """
    x = np.ascontiguousarray(x).astype(np.int32)
    y = np.ascontiguousarray(y).astype(np.int32)
    x_dev = cuda.to_device(x)
    y_dev = cuda.to_device(y)
    results = cuda.device_array((x.shape[0]), dtype=np.int32)
    bpg = (x.shape[0] + (THREADS_PER_BLOCK - 1)) // THREADS_PER_BLOCK
    _cuda_direction_from_two_bps[bpg, THREADS_PER_BLOCK](x_dev, y_dev, results)
    results = results.copy_to_host()
    return results


def sliding_circular_hotspots(x: np.ndarray,
                              time_window: float,
                              sample_rate: float,
                              bins: np.ndarray,
                              batch_size: Optional[int] = int(3.5e+7)) -> np.ndarray:
    """
    Calculate the proportion of data points falling within specified circular bins over a sliding time window using GPU

    This function processes time series data representing angles (in degrees) and calculates the proportion of data
    points within specified angular bins over a sliding window. The calculations are performed in batches to
    accommodate large datasets efficiently.

    .. csv-table::
       :header: EXPECTED RUNTIMES
       :file: ../../../docs/tables/sliding_circular_hotspots.csv
       :widths: 10, 45, 45
       :align: center
       :header-rows: 1

    .. seealso::
       For CPU function see :func:`~simba.mixins.circular_statistics.CircularStatisticsMixin.sliding_circular_hotspots`.


    :param np.ndarray x: The input time series data in degrees. Should be a 1D numpy array.
    :param float time_window: The size of the sliding window in seconds.
    :param float sample_rate: The sample rate of the time series data (i.e., hz, fps).
    :param ndarray bins: 2D array of shape representing circular bins defining [start_degree, end_degree] inclusive.
    :param Optional[int] batch_size: The size of each batch for processing the data. Default is 5e+7 (50m).
    :return: A 2D numpy array where each row corresponds to a time point in `data`, and each column represents a circular bin. The values in the array represent the proportion of data points within each bin at each time point. The first column represents the first bin.
    :rtype: np.ndarray
    """

    n = x.shape[0]
    x = cp.asarray(x, dtype=cp.float16)
    results = cp.full((x.shape[0], bins.shape[0]), dtype=cp.float16, fill_value=-1)
    window_size = int(cp.ceil(time_window * sample_rate))
    for cnt, left in enumerate(range(0, n, batch_size)):
        right = int(min(left + batch_size, n))
        if cnt > 0:
            left = left - window_size + 1
        x_batch = x[left:right]
        x_batch = cp.lib.stride_tricks.sliding_window_view(x_batch, window_size).astype(cp.float16)
        batch_results = cp.full((x_batch.shape[0], bins.shape[0]), dtype=cp.float16, fill_value=-1)
        for bin_cnt in range(bins.shape[0]):
            if bins[bin_cnt][0] > bins[bin_cnt][1]:
                mask = ((x_batch >= bins[bin_cnt][0]) & (x_batch <= 360)) | ((x_batch >= 0) & (x_batch <= bins[bin_cnt][1]))
            else:
                mask = (x_batch >= bins[bin_cnt][0]) & (x_batch <= bins[bin_cnt][1])
            count_per_row = cp.array(mask.sum(axis=1) / window_size).reshape(-1, )
            batch_results[:, bin_cnt] = count_per_row
        results[left + window_size - 1:right, ] = batch_results
    return results.get()

def sliding_circular_mean(x: np.ndarray,
                          time_window: float,
                          sample_rate: int,
                          batch_size: Optional[int] = 3e+7) -> np.ndarray:

    """
    Calculate the sliding circular mean over a time window for a series of angles.

    This function computes the circular mean of angles in the input array `x` over a specified sliding window.
    The circular mean is a measure of the average direction for angles, which is especially useful for angular data
    where traditional averaging would not be meaningful due to the circular nature of angles (e.g., 359° and 1° should average to 0°).

    The calculation is performed using a sliding window approach, where the circular mean is computed for each window
    of angles. The function leverages GPU acceleration via CuPy for efficiency when processing large datasets.

    The circular mean :math:`\\mu` for a set of angles is calculated using the following formula:

    .. math::

        \\mu = \\text{atan2}\\left(\\frac{1}{N} \\sum_{i=1}^{N} \\sin(\\theta_i), \\frac{1}{N} \\sum_{i=1}^{N} \\cos(\\theta_i)\\right)

    - :math:`\\theta_i` are the angles in radians within the sliding window
    - :math:`N` is the number of samples in the window


    .. csv-table::
       :header: EXPECTED RUNTIMES
       :file: ../../../docs/tables/sliding_circular_mean.csv
       :widths: 10, 45, 45
       :align: center
       :header-rows: 1

    .. seealso::
       For CPU function see :func:`~simba.mixins.circular_statistics.CircularStatisticsMixin.sliding_circular_mean`.

    :param np.ndarray x: Input array containing angle values in degrees. The array should be 1-dimensional.
    :param float time_window: Time duration for the sliding window, in seconds. This determines the number of samples in each window  based on the `sample_rate`.
    :param int sample_rate: The number of samples per second (i.e., FPS). This is used to calculate the window size in terms of array indices.
    :param Optional[int] batch_size: The maximum number of elements to process in each batch. This is used to handle large arrays by processing them in chunks to avoid memory overflow. Defaults to 3e+7 (30 million elements).
    :return np.ndarray: A 1D numpy array of the same length as `x`, containing the circular mean for each sliding window.  Values before the window is fully populated will be set to -1.

    :example:
    >>> x = np.random.randint(0, 361, (i, )).astype(np.int32)
    >>> results = sliding_circular_mean(x, 1, 10)
    """


    window_size = np.ceil(time_window * sample_rate).astype(np.int64)
    n = x.shape[0]
    results = cp.full(x.shape[0], -1, dtype=np.int32)
    for cnt, left in enumerate(range(0, int(n), int(batch_size))):
        right = np.int32(min(left + batch_size, n))
        if cnt > 0:
            left = left - window_size+1
        x_batch = cp.asarray(x[left:right])
        x_batch = cp.lib.stride_tricks.sliding_window_view(x_batch, window_size)
        x_batch = np.deg2rad(x_batch)
        cos, sin = cp.cos(x_batch).astype(np.float32), cp.sin(x_batch).astype(np.float32)
        r = cp.rad2deg(cp.arctan2(cp.mean(sin, axis=1), cp.mean(cos, axis=1)))
        r = cp.where(r < 0, r + 360, r)
        results[left + window_size - 1:right] = r
    return results.get()



def sliding_circular_range(x: np.ndarray,
                          time_window: float,
                          sample_rate: float,
                          batch_size: Optional[int] = int(5e+7)) -> np.ndarray:
    """
    Computes the sliding circular range of a time series data array using GPU.

    This function calculates the circular range of a time series data array using a sliding window approach.
    The input data is assumed to be in degrees, and the function handles the circular nature of the data
    by considering the circular distance between angles.

    .. math::

       R = \\min \\left( \\text{max}(\\Delta \\theta) - \\text{min}(\\Delta \\theta), \\, 360 - \\text{max}(\\Delta \\theta) + \\text{min}(\\Delta \\theta) \\right)

    where:

    - :math:`\\Delta \\theta` is the difference between angles within the window,
    - :math:`360` accounts for the circular nature of the data (i.e., wrap-around at 360 degrees).

    .. csv-table::
       :header: EXPECTED RUNTIMES
       :file: ../../../docs/tables/sliding_circular_range.csv
       :widths: 10, 45, 45
       :align: center
       :header-rows: 1

    .. seealso::
       For CPU function see :func:`~simba.mixins.circular_statistics.CircularStatisticsMixin.sliding_circular_range`.

    :param np.ndarray x: The input time series data in degrees. Should be a 1D numpy array.
    :param float time_window: The size of the sliding window in seconds.
    :param float sample_rate: The sample rate of the time series data (i.e., hz, fps).
    :param Optional[int] batch_size: The size of each batch for processing the data. Default is 5e+7 (50m).
    :return: A numpy array containing the sliding circular range values.
    :rtype: np.ndarray

    :example:
    >>> x = np.random.randint(0, 361, (19, )).astype(np.int32)
    >>> p = sliding_circular_range(x, 1, 10)
    """

    n = x.shape[0]
    x = cp.asarray(x, dtype=cp.float16)
    results = cp.zeros_like(x, dtype=cp.int16)
    x = cp.deg2rad(x).astype(cp.float16)
    window_size = int(cp.ceil(time_window * sample_rate))
    for cnt, left in enumerate(range(0, n, batch_size)):
        right = int(min(left + batch_size, n))
        if cnt > 0:
            left = left - window_size + 1
        x_batch = x[left:right]
        x_batch = cp.lib.stride_tricks.sliding_window_view(x_batch, window_size).astype(cp.float16)
        x_batch = cp.sort(x_batch)
        results[left + window_size - 1:right] = cp.abs(cp.rint(cp.rad2deg(cp.amin(cp.vstack([x_batch[:, -1] - x_batch[:, 0], 2 * cp.pi - cp.max(cp.diff(x_batch), axis=1)]).T, axis=1))))
    return results.get()




def sliding_circular_std(x: np.ndarray,
                         time_window: float,
                         sample_rate: float,
                         batch_size: Optional[int] = int(5e+7)) -> np.ndarray:

    r"""
    Calculate the sliding circular standard deviation of a time series data on GPU.

    This function computes the circular standard deviation over a sliding window for a given time series array.
    The time series data is assumed to be in degrees, and the function converts it to radians for computation.
    The sliding window approach is used to handle large datasets efficiently, processing the data in batches.

    The circular standard deviation (σ) is computed using the formula:

    .. math::

       \sigma = \sqrt{-2 \cdot \log \left|\text{mean}\left(\exp(i \cdot x_{\text{batch}})\right)\right|}

    where :math:`x_{\text{batch}}` is the data within the current sliding window, and :math:`\text{mean}` and
    :math:`\log` are computed in the circular (complex plane) domain.

    .. csv-table::
       :header: EXPECTED RUNTIMES
       :file: ../../../docs/tables/sliding_circular_std.csv
       :widths: 10, 45, 45
       :align: center
       :header-rows: 1

    .. seealso::
       For CPU function see :func:`~simba.mixins.circular_statistics.CircularStatisticsMixin.sliding_circular_std`.

    :param np.ndarray x: The input time series data in degrees. Should be a 1D numpy array.
    :param float time_window: The size of the sliding window in seconds.
    :param float sample_rate: The sample rate of the time series data (i.e., hz, fps).
    :param Optional[int] batch_size: The size of each batch for processing the data. Default is 5e+7 (50m).

    :return: A numpy array containing the sliding circular standard deviation values.
    :rtype: np.ndarray
    """


    n = x.shape[0]
    x = cp.asarray(x, dtype=cp.float16)
    results = cp.zeros_like(x, dtype=cp.float16)
    x = np.deg2rad(x).astype(cp.float16)
    window_size = int(np.ceil(time_window * sample_rate))
    for cnt, left in enumerate(range(0, n, batch_size)):
        right = int(min(left + batch_size, n))
        if cnt > 0:
            left = left - window_size + 1
        x_batch = x[left:right]
        x_batch = cp.lib.stride_tricks.sliding_window_view(x_batch, window_size).astype(cp.float16)
        m = cp.log(cp.abs(cp.mean(cp.exp(1j * x_batch), axis=1)))
        stdev = cp.rad2deg(cp.sqrt(-2 * m))
        results[left + window_size - 1:right] = stdev

    return results.get()


def sliding_rayleigh_z(x: np.ndarray,
                       time_window: float,
                       sample_rate: float,
                       batch_size: Optional[int] = int(5e+7)) -> Tuple[np.ndarray, np.ndarray]:

    r"""
    Computes the Rayleigh Z-statistic over a sliding window for a given time series of angles

    This function calculates the Rayleigh Z-statistic, which tests the null hypothesis that the population of angles
    is uniformly distributed around the circle. The calculation is performed over a sliding window across the input
    time series, and results are computed in batches for memory efficiency.

    Data is processed using GPU acceleration via CuPy, which allows for faster computation compared to a CPU-based approach.

    .. note::
        Adapted from ``pingouin.circular.circ_rayleigh`` and ``pycircstat.tests.rayleigh``.


    **Rayleigh Z-statistic:**

    The Rayleigh Z-statistic is given by:

    .. math::

       R = \frac{1}{n} \sqrt{\left(\sum_{i=1}^{n} \cos(\theta_i)\right)^2 + \left(\sum_{i=1}^{n} \sin(\theta_i)\right)^2}

    where:
    - :math:`\theta_i` are the angles in the window.
    - :math:`n` is the number of angles in the window.


    .. csv-table::
       :header: EXPECTED RUNTIMES
       :file: ../../../docs/tables/sliding_rayleigh_z.csv
       :widths: 10, 45, 45
       :align: center
       :header-rows: 1

    .. seealso::
       For CPU function see :func:`~simba.mixins.circular_statistics.CircularStatisticsMixin.sliding_rayleigh_z`.


    :param np.ndarray x: Input array of angles in degrees. Should be a 1D numpy array.
    :param float time_window: The size of the sliding window in time units (e.g., seconds).
    :param float sample_rate: The sampling rate of the input time series in samples per time unit (e.g., Hz, fps).
    :param Optional[int] batch_size: The number of samples to process in each batch. Default is 5e7 (50m). Reducing this value may save memory at the cost of longer computation time.
    :return:
       A tuple containing two numpy arrays:
       - **z_results**: Rayleigh Z-statistics for each position in the input array where the window was fully applied.
       - **p_results**: Corresponding p-values for the Rayleigh Z-statistics.
    :rtype: Tuple[np.ndarray, np.ndarray]
    """

    n = x.shape[0]
    x = cp.asarray(x, dtype=cp.float16)
    z_results = cp.zeros_like(x, dtype=cp.float16)
    p_results = cp.zeros_like(x, dtype=cp.float16)
    x = np.deg2rad(x).astype(cp.float16)
    window_size = int(np.ceil(time_window * sample_rate))
    for cnt, left in enumerate(range(0, n, batch_size)):
        right = int(min(left + batch_size, n))
        if cnt > 0:
            left = left - window_size + 1
        x_batch = x[left:right]
        x_batch = cp.lib.stride_tricks.sliding_window_view(x_batch, window_size).astype(cp.float16)
        cos_sums = cp.nansum(cp.cos(x_batch), axis=1) ** 2
        sin_sums = cp.nansum(cp.sin(x_batch), axis=1) ** 2
        R = cp.sqrt(cos_sums + sin_sums) / window_size
        Z = window_size * (R**2)
        P = cp.exp(np.sqrt(1 + 4 * window_size + 4 * (window_size ** 2 - R ** 2)) - (1 + 2 * window_size))
        z_results[left + window_size - 1:right] = Z
        p_results[left + window_size - 1:right] = P

    return z_results.get(), p_results.get()


def sliding_resultant_vector_length(x: np.ndarray,
                                    time_window: float,
                                    sample_rate: int,
                                    batch_size: Optional[int] = 3e+7) -> np.ndarray:

    """
    Calculate the sliding resultant vector length over a time window for a series of angles.

    This function computes the resultant vector length (R) for each window of angles in the input array `x`.
    The resultant vector length is a measure of the concentration of angles, and it ranges from 0 to 1, where 1
    indicates all angles point in the same direction, and 0 indicates uniform distribution of angles.

    For a given sliding window of angles, the resultant vector length :math:`R` is calculated using the following formula:

    .. math::

        R = \\frac{1}{N} \\sqrt{\\left(\\sum_{i=1}^{N} \\cos(\\theta_i)\\right)^2 + \\left(\\sum_{i=1}^{N} \\sin(\\theta_i)\\right)^2}

    where:

    - :math:`\\theta_i` are the angles in radians within the sliding window
    - :math:`N` is the number of samples in the window

    The computation is performed in a sliding window manner over the entire array, utilizing GPU acceleration
    with CuPy for efficiency, especially on large datasets.


    .. csv-table::
       :header: EXPECTED RUNTIMES
       :file: ../../../docs/tables/sliding_resultant_vector_length.csv
       :widths: 10, 10, 80
       :align: center
       :header-rows: 1

    .. seealso::
       For CPU function see :func:`~simba.mixins.circular_statistics.CircularStatisticsMixin.sliding_resultant_vector_length`.

    :param np.ndarray x: Input array containing angle values in degrees. The array should be 1-dimensional.
    :param float time_window: Time duration for the sliding window, in seconds. This determines the number of samples in each window  based on the `sample_rate`.
    :param int sample_rate: The number of samples per second (i.e., FPS). This is used to calculate the window size in terms of array indices.
    :param Optional[int] batch_size: The maximum number of elements to process in each batch. This is used to handle large arrays by processing them in chunks to avoid memory overflow. Defaults to 3e+7 (30 million elements).
    :return np.ndarray: A 1D numpy array of the same length as `x`, containing the resultant vector length for each sliding window. Values before the window is fully populated will be set to -1.

    :example:
    >>> x = np.random.randint(0, 361, (5000, )).astype(np.int32)
    >>> results = sliding_resultant_vector_length(x, 1, 10)
    """

    window_size = np.ceil(time_window * sample_rate).astype(np.int64)
    n = x.shape[0]
    results = cp.full(x.shape[0], -1, dtype=np.float32)
    for cnt, left in enumerate(range(0, int(n), int(batch_size))):
        right = np.int32(min(left + batch_size, n))
        if cnt > 0:
            left = left - window_size+1
        x_batch = cp.asarray(x[left:right])
        x_batch = cp.lib.stride_tricks.sliding_window_view(x_batch, window_size)
        x_batch = np.deg2rad(x_batch)
        cos, sin = cp.cos(x_batch).astype(np.float32), cp.sin(x_batch).astype(np.float32)
        cos_sum, sin_sum = cp.sum(cos, axis=1), cp.sum(sin, axis=1)
        r = np.sqrt(cos_sum ** 2 + sin_sum ** 2) / window_size
        results[left+window_size-1:right] = r
    return results.get()


def direction_from_three_bps(x: np.ndarray,
                             y: np.ndarray,
                             z: np.ndarray,
                             batch_size: Optional[int] = int(1.5e+7)) -> np.ndarray:

    """
    Calculate the direction angle based on the coordinates of three body points using GPU acceleration.

    This function computes the mean direction angle (in degrees) for a batch of coordinates
    provided in the form of NumPy arrays. The calculation is based on the arctangent of the
    difference in x and y coordinates between pairs of points. The result is a value in
    the range [0, 360) degrees.

    .. seealso::
       :func:`simba.mixins.circular_statistics.CircularStatisticsMixin.direction_three_bps`

    :param np.ndarray x: A 2D array of shape (N, 2) containing the x-coordinates of the first body part  (nose)
    :param np.ndarray y: A 2D array of shape (N, 2) containing the coordinates of the second body part (left ear).
    :param np.ndarray z: A 2D array of shape (N, 2) containing the coordinates of the second body part (right ear).
    :param Optional[int] batch_size: The size of the batch to be processed in each iteration. Default is 15 million.
    :return: An array of shape (N,) containing the computed direction angles in degrees.
    :rtype np.ndarray:
    """

    check_valid_array(data=x, source=direction_from_three_bps.__name__, accepted_ndims=(2,), accepted_dtypes=Formats.NUMERIC_DTYPES.value)
    check_valid_array(data=y, source=direction_from_three_bps.__name__, accepted_shapes=(x.shape,), accepted_dtypes=Formats.NUMERIC_DTYPES.value)
    check_valid_array(data=z, source=direction_from_three_bps.__name__, accepted_shapes=(x.shape,), accepted_dtypes=Formats.NUMERIC_DTYPES.value)
    check_int(value=batch_size, name=direction_from_three_bps.__name__, min_value=1)
    results = cp.full((x.shape[0]), fill_value=-1, dtype=np.int16)

    for l in range(0, x.shape[0], batch_size):
        r = l + batch_size
        x_batch = cp.array(x[l:r])
        y_batch = cp.array(y[l:r])
        z_batch = cp.array(z[l:r])
        left_ear_to_nose = cp.arctan2(x_batch[:, 0] - y_batch[:, 0], y_batch[:, 1] - x_batch[:,1])
        right_ear_nose = cp.arctan2(x_batch[:, 0] - z_batch[:, 0], z_batch[:, 1] - x_batch[:, 1])
        mean_angle_rad = cp.arctan2(cp.sin(left_ear_to_nose) + cp.sin(right_ear_nose), cp.cos(left_ear_to_nose) + cp.cos(right_ear_nose))
        results[l:r] = (cp.degrees(mean_angle_rad) + 360) % 360

    return results.get()


@cuda.jit()
def _instantaneous_angular_velocity(x, stride, results):
    r = cuda.grid(1)
    l = np.int32(r - (stride[0]))
    if (r > results.shape[0]) or (l < 0):
        results[r] = -1
    else:
        d = math.pi - (abs(math.pi - abs(x[l] - x[r])))
        results[r] = d * (180 / math.pi)


def instantaneous_angular_velocity(x: np.ndarray, stride: Optional[int] = 1) -> np.ndarray:
    r"""
    Calculate the instantaneous angular velocity between angles in a given array.

    This function uses CUDA to perform parallel computations on the GPU.

    The angular velocity is computed using the difference in angles between
    the current and previous values (with a specified stride) in the array.
    The result is returned in degrees per unit time.

    .. csv-table::
       :header: EXPECTED RUNTIMES
       :file: ../../../docs/tables/instantaneous_angular_velocity.csv
       :widths: 10, 90
       :align: center
       :header-rows: 1

    .. math::
       \omega = \frac{{\Delta \theta}}{{\Delta t}} = \frac{{180}}{{\pi}} \times \left( \pi - \left| \pi - \left| \theta_r - \theta_l \right| \right| \right)

    where:
    - \( \theta_r \) is the current angle.
    - \( \theta_l \) is the angle at the specified stride before the current angle.
    - \( \Delta t \) is the time difference between the two angles.


    .. seealso::
       :func:`simba.mixins.circular_statistics.CircularStatisticsMixin.instantaneous_angular_velocity`

    :param np.ndarray x: Array of angles in degrees, for which the instantaneous angular velocity will be calculated.
    :param Optional[int] stride: The stride or lag (in frames) to use when calculating the difference in angles. Defaults to 1.
    :return: Array of instantaneous angular velocities corresponding to the input angles. Velocities are in degrees per unit time.
    :rtype: np.ndarray
    """

    x = np.deg2rad(x).astype(np.int16)
    stride = np.array([stride]).astype(np.int64)
    bpg = (x.shape[0] + (THREADS_PER_BLOCK - 1)) // THREADS_PER_BLOCK
    x_dev = cuda.to_device(x)
    stride_dev = cuda.to_device(stride)
    results = cuda.device_array(x.shape[0], dtype=np.float32)
    _instantaneous_angular_velocity[bpg, THREADS_PER_BLOCK](x_dev, stride_dev, results)
    return results.copy_to_host()


@cuda.jit(device=True)
def _rad2deg(x):
    return x * (180/math.pi)


@cuda.jit()
def _sliding_bearing(x, stride, results):
    r = cuda.grid(1)
    l = np.int32(r - (stride[0]))
    if (r > results.shape[0]-1) or (l < 0):
        results[r] = -1
    else:
        x1, y1 = x[l, 0], x[l, 1]
        x2, y2 = x[r, 0], x[r, 1]
        bearing = _rad2deg(math.atan2(x2 - x1, y2 - y1))
        results[r] = (bearing + 360) % 360


def sliding_bearing(x: np.ndarray,
                    stride: Optional[float] = 1,
                    sample_rate: Optional[float] = 1) -> np.ndarray:
    """
    Compute the bearing between consecutive points in a 2D coordinate array using a sliding window approach using GPU acceleration.

    This function calculates the angle (bearing) in degrees between each point and a point a certain number of
    steps ahead (defined by `stride`) in the 2D coordinate array `x`. The bearing is calculated using the
    arctangent of the difference in coordinates, converted from radians to degrees.

    .. csv-table::
       :header: EXPECTED RUNTIMES
       :file: ../../../docs/tables/sliding_bearing.csv
       :widths: 10, 45, 45
       :align: center
       :header-rows: 1

    .. seealso::
       :func:`simba.mixins.circular_statistics.CircularStatisticsMixin.sliding_bearing`

    :param np.ndarray x: A 2D array of shape `(n, 2)` where each row represents a point with `x` and `y` coordinates. The array must be numeric.
    :param Optional[float] stride: The time (multiplied by `sample_rate`) to look ahead when computing the bearing in seconds. Defaults to 1.
    :param Optional[float] sample_rate: A multiplier applied to the `stride` value to determine the actual step size for calculating the bearing. E.g., frames per second. Defaults to 1. If the resulting stride is less than 1, it is automatically set to 1.
    :return:A 1D array of shape `(n,)` containing the calculated bearings in degrees. Values outside the valid range (i.e., where the stride exceeds array bounds) are set to -1.
    :rtype: np.ndarray
    """

    check_valid_array(data=x, source=f'{sliding_bearing.__name__} x', accepted_ndims=(2,), accepted_axis_1_shape=(2,), accepted_dtypes=Formats.NUMERIC_DTYPES.value)
    check_float(name=f'{sliding_bearing.__name__} stride', value=stride, min_value=10e-6, max_value=x.shape[0]-1)
    check_float(name=f'{sliding_bearing.__name__} sample_rate', value=sample_rate, min_value=10e-6, max_value=x.shape[0]-1)
    stride = int(stride * sample_rate)
    if stride < 1:
        stride = 1
    stride = np.array([stride]).astype(np.int64)
    bpg = (x.shape[0] + (THREADS_PER_BLOCK - 1)) // THREADS_PER_BLOCK
    x_dev = cuda.to_device(x)
    stride_dev = cuda.to_device(stride)
    results = cuda.device_array(x.shape[0], dtype=np.float32)
    _sliding_bearing[bpg, THREADS_PER_BLOCK](x_dev, stride_dev, results)
    return results.copy_to_host()


@cuda.jit(device=True)
def _rad2deg(x):
    return x * (180 / math.pi)


@cuda.jit()
def _sliding_angular_diff(data, strides, results):
    x, y = cuda.grid(2)
    if (x > data.shape[0] - 1) or (y > strides.shape[0] - 1):
        return
    else:
        stride = int(strides[y])
        if x - stride < 0:
            return
        a_2 = data[x]
        a_1 = data[x - stride]
        distance = math.pi - abs(math.pi - abs(a_1 - a_2))
        distance = abs(int(_rad2deg(distance)) + 1)
        results[x][y] = distance


def sliding_angular_diff(x: np.ndarray,
                         time_windows: np.ndarray,
                         fps: float) -> np.ndarray:
    """
    Calculate the sliding angular differences for a given time window using GPU acceleration.


    This function computes the angular differences between each angle in `x`
    and the corresponding angle located at a distance determined by the time window
    and frame rate (fps). The results are returned as a 2D array where each row corresponds
    to a position in `x`, and each column corresponds to a different time window.

    .. csv-table::
       :header: EXPECTED RUNTIMES
       :file: ../../../docs/tables/sliding_angular_diff.csv
       :widths: 10, 45, 45
       :align: center
       :header-rows: 1


    .. seealso::
       :func:`simba.mixins.circular_statistics.CircularStatisticsMixin.sliding_angular_diff`

    .. math::
       \text{difference} = \pi - |\pi - |a_1 - a_2||

    Where:
    - :math:`a_1` is the angle at position `x`.
    - :math:`a_2` is the angle at position `x - \text{stride}`.

    :param np.ndarray x: 1D array of angles in degrees.
    :param np.ndarray time_windows: 1D array of time windows in seconds to determine the stride (distance in frames) between angles.
    :param float fps: Frame rate (frames per second) used to convert time windows to strides.
    :return: 2D array of angular differences. Each row corresponds to an angle in `x`, and each column corresponds to a time window.
    :rtype: np.ndarray
    """

    x = np.deg2rad(x)
    strides = np.zeros(time_windows.shape[0])
    for i in range(time_windows.shape[0]):
        strides[i] = np.ceil(time_windows[i] * fps).astype(np.int32)
    x_dev = cuda.to_device(x)
    stride_dev = cuda.to_device(strides)
    results = cuda.device_array((x.shape[0], time_windows.shape[0]))
    grid_x = (x.shape[0] + THREADS_PER_BLOCK - 1) // THREADS_PER_BLOCK
    grid_y = (strides.shape[0] + THREADS_PER_BLOCK - 1)
    blocks_per_grid = (grid_x, grid_y)
    _sliding_angular_diff[blocks_per_grid, THREADS_PER_BLOCK](x_dev, stride_dev, results)
    results = results.copy_to_host().astype(np.int32)
    return results

@cuda.jit()
def _rotational_direction(data, stride, results):
    r = cuda.grid(1)
    l = int(r - stride[0])
    if (r < 0) or (r > data.shape[0] - 1):
        return
    elif (l < 0):
        return
    else:
        l_val, r_val = data[l], data[r]
        angle_diff = r_val - l_val
        if angle_diff > np.pi:
            angle_diff -= 2 * np.pi
        elif angle_diff < -np.pi:
            angle_diff += 2 * np.pi
        if angle_diff == 0:
            results[r] = 0
        elif angle_diff > 0:
            results[r] = 1
        else:
            results[r] = 2

def rotational_direction(data: np.ndarray, stride: Optional[int] = 1) -> np.ndarray:
    """
    Computes the rotational direction between consecutive data points in a circular space, where the angles wrap
    around at 360 degrees. The function uses GPU acceleration via CUDA to process the data in parallel.

    The result array contains values:

    * `0` where there is no change between points.
    * `1` where the angle has increased in the positive direction.
    * `2` where the angle has decreased in the negative direction.

    .. seealso::
       :func:`simba.mixins.circular_statistics.CircularStatisticsMixin.rotational_direction` for jitted CPU method.

    :param np.ndarray data: 1D array of angular data (in degrees) to analyze. The data will be internally converted to radians and wrapped between [0, 360) degrees before processing.
    :param Optional[int] stride: The stride or gap between data points for which the rotational direction is calculated. Default is 1.
    :return: A 1D array of integers of the same length as `data`, where each element indicates the rotational direction between the current and previous point based on the stride. The first `stride` elements in the result will  be initialized to -1 since they cannot be compared.
    :rtype: np.ndarray

    :example:
    >>> data = np.random.randint(0, 365, (100))
    >>> p = rotational_direction(data=data)
    """
    data = np.deg2rad(data % 360)
    results = np.full((data.shape[0]), fill_value=-1, dtype=np.int16)
    results_dev = cuda.to_device(results)
    stride = np.array([stride])
    stride_dev = cuda.to_device(stride)
    data_dev = cuda.to_device(data)
    bpg = (data.shape[0] + (THREADS_PER_BLOCK - 1)) // THREADS_PER_BLOCK
    _rotational_direction[bpg, THREADS_PER_BLOCK](data_dev, stride_dev, results_dev)
    return results_dev.copy_to_host()






__author__ = "Simon Nilsson"

import glob
import os
from copy import deepcopy
from typing import Dict, Optional, Union

import numpy as np
import pandas as pd

from simba.mixins.config_reader import ConfigReader
from simba.utils.checks import (
    check_all_file_names_are_represented_in_video_log, check_if_dir_exists,
    check_if_filepath_list_is_empty)
from simba.utils.errors import NoFilesFoundError
from simba.utils.printing import stdout_success
from simba.utils.read_write import (bento_file_reader,
                                    find_files_of_filetypes_in_directory,
                                    get_fn_ext, read_df, write_df)
from simba.utils.warnings import (
    ThirdPartyAnnotationsClfMissingWarning,
    ThirdPartyAnnotationsOutsidePoseEstimationDataWarning)


class BentoAppender(ConfigReader):
    """
    Append BENTO annotation to SimBA featurized datasets.

    .. note::
       `Example BENTO input file <https://github.com/sgoldenlab/simba/blob/master/misc/bento_example.annot>`_.
       'GitHub tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/third_party_annot.md>`_.

    :param str config_path: path to SimBA project config file in Configparser format
    :param str data_dir: Path to folder containing BENTO data.

    :example:
    >>> bento_dir = 'tests/test_data/bento_example'
    >>> config_path = 'tests/test_data/import_tests/project_folder/project_config.ini'
    >>> bento_appender = BentoAppender(config_path=config_path, data_dir=bento_dir)
    >>> bento_appender.run()

    References
    ----------
    .. [1] Segalin et al., eLife, https://doi.org/10.7554/eLife.63720
    """

    def __init__(self,
                 config_path: Union[str, os.PathLike],
                 data_dir: Union[str, os.PathLike]):

        ConfigReader.__init__(self, config_path=config_path)
        check_if_dir_exists(in_dir=data_dir)
        self.bento_files = find_files_of_filetypes_in_directory(directory=data_dir, extensions=['.annot'], raise_error=False, raise_warning=True)
        check_if_filepath_list_is_empty(filepaths=self.feature_file_paths, error_msg="SIMBA ERROR: No feature files found in project_folder/csv/features_extracted. Extract Features BEFORE appending BENTO annotations")
        check_if_filepath_list_is_empty(filepaths=self.bento_files, error_msg=f"SIMBA ERROR: No BENTO files with .annot extension found in {data_dir}.")
        self.saved_files = []

    def run(self):
        check_all_file_names_are_represented_in_video_log(video_info_df=self.video_info_df, data_paths=self.bento_files)
        for file_cnt, bento_file_path in enumerate(self.bento_files):
            _, self.video_name, ext = get_fn_ext(filepath=bento_file_path)
            print(f"Appending BENTO annotation to video {self.video_name}...")
            _, _, fps = self.read_video_info(video_name=self.video_name)
            features_path = os.path.join(self.features_dir, self.video_name + f'.{self.file_type}')
            if not os.path.isfile(features_path):
                raise NoFilesFoundError(msg=f'No features file for annotation file {self.video_name} file in {self.features_dir}. SimBA is expecting a file at path {features_path}')
            self.save_path = os.path.join(self.targets_folder, self.video_name + f'.{self.file_type}')
            feature_df = read_df(file_path=features_path, file_type=self.file_type)
            self.results = deepcopy(feature_df)
            bento_dict = bento_file_reader(file_path=bento_file_path, fps=fps, save_path=None, orient='index')
            for clf_name in self.clf_names:
                self.results[clf_name] = 0
                if clf_name not in bento_dict.keys():
                    ThirdPartyAnnotationsClfMissingWarning(video_name=self.video_name, clf_name=clf_name)
                else:
                    clf_bento_df = bento_dict[clf_name]
                    annotations_idx = [i for s in list(clf_bento_df.apply(lambda x: list(range(int(x["START"]), int(x["STOP"]))), 1)) for i in s]
                    annotations_idx_outside_video = [x for x in annotations_idx if x > len(feature_df)]
                    valid_annotation_ids = [x for x in annotations_idx if x < len(feature_df)]

                    if len(annotations_idx_outside_video) > 0:
                        ThirdPartyAnnotationsOutsidePoseEstimationDataWarning(video_name=self.video_name,
                                                                              clf_name=clf_name,
                                                                              frm_cnt=len(feature_df),
                                                                              first_error_frm=annotations_idx_outside_video[0],
                                                                              ambiguous_cnt=len(annotations_idx_outside_video))
                    if len(valid_annotation_ids) > 0:
                        print(f"Appending {str(len(valid_annotation_ids))} {clf_name} frame annotations to video {self.video_name}...")
                        self.results.loc[valid_annotation_ids, clf_name] = 1
            self.__save()
            stdout_success(msg=f"Annotations for {str(len(self.saved_files))} video(s) and saved in the {self.targets_folder}.")

    def __save(self):
        write_df(df=self.results, file_type=self.file_type, save_path=self.save_path)
        self.saved_files.append(self.save_path)
        print(f"BENTO annotations appended to video {self.video_name} and saved in {self.save_path}")



#
#
#     #
#     #
#     #         annotation_df = pd.read_csv(
#     #             bento_path, delim_whitespace=True, index_col=False, low_memory=False
#     #         )
#     #         start_idx = annotation_df.index[
#     #             annotation_df["Bento"] == "Ch1----------"
#     #         ].values[0]
#     #         sliced_annot = annotation_df.iloc[start_idx + 1 :]
#     #         annotated_behaviors = sliced_annot[sliced_annot["Bento"].str.contains(">")][
#     #             "Bento"
#     #         ].tolist()
#     #         annotated_behavior_names = [x[1:] for x in annotated_behaviors]
#     #         missing_annotation = set(self.clf_names) - set(annotated_behavior_names)
#     #         missing_clf = list(set(annotated_behavior_names) - set(self.clf_names))
#     #         annotation_intersection = [
#     #             x for x in self.clf_names if x in annotated_behavior_names
#     #         ]
#     #         for missing_clf in missing_annotation:
#     #             ThirdPartyAnnotationsClfMissingWarning(
#     #                 video_name=self.video_name, clf_name=missing_clf
#     #             )
#     #             self.results_df[missing_clf] = 0
#     #         if missing_clf:
#     #             ThirdPartyAnnotationsAdditionalClfWarning(
#     #                 video_name=self.video_name, clf_names=missing_clf
#     #             )
#     #
#     #         for clf_name in annotation_intersection:
#     #             self.results_df[clf_name] = 0
#     #             clf_start_idx = sliced_annot.index[
#     #                 sliced_annot["Bento"] == f">{clf_name}"
#     #             ].values[0]
#     #             clf_df = sliced_annot.loc[clf_start_idx + 2 :, :]
#     #             end_idx = (
#     #                 clf_df.isnull()[clf_df.isnull().any(axis=1)].idxmax(axis=1).index
#     #             )
#     #             if end_idx.values:
#     #                 end_idx = end_idx.values[0]
#     #             else:
#     #                 end_idx = max(clf_df.index + 1)
#     #             clf_df = clf_df.loc[: end_idx - 1, :].reset_index(drop=True)
#     #             clf_df.columns = ["start_time", "stop_time", "duration"]
#     #             clf_df["start_frm"] = clf_df["start_time"].astype(float) * fps
#     #             clf_df["end_frm"] = clf_df["stop_time"].astype(float) * fps
#     #             clf_df["start_frm"] = clf_df["start_frm"].astype(int)
#     #             clf_df["end_frm"] = clf_df["end_frm"].astype(int)
#     #             annotations_idx = list(
#     #                 clf_df.apply(
#     #                     lambda x: list(
#     #                         range(int(x["start_frm"]), int(x["end_frm"]) + 1)
#     #                     ),
#     #                     1,
#     #                 )
#     #             )
#     #             annotations_idx = [i for s in annotations_idx for i in s]
#     #             annotations_idx_outside_video = [
#     #                 x for x in annotations_idx if x > video_frm_length
#     #             ]
#     #             valid_annotation_ids = [
#     #                 x for x in annotations_idx if x <= video_frm_length
#     #             ]
#     #             if len(annotations_idx_outside_video):
#     #                 ThirdPartyAnnotationsOutsidePoseEstimationDataWarning(
#     #                     video_name=self.video_name,
#     #                     clf_name=clf_name,
#     #                     frm_cnt=len(feature_df),
#     #                     first_error_frm=annotations_idx_outside_video[0],
#     #                     ambiguous_cnt=len(annotations_idx_outside_video),
#     #                 )
#     #             if len(valid_annotation_ids) > 0:
#     #                 print(
#     #                     f"Appending {str(len(valid_annotation_ids))} {clf_name} frame annotations to video {self.video_name}..."
#     #                 )
#     #                 self.results_df.loc[valid_annotation_ids, clf_name] = 1
#     #         self.__save()
#     #     stdout_success(
#     #         msg=f"Annotations for {str(len(self.saved_files))} video(s) and saved in project_folder/csv/targets_inserted directory."
#     #     )
#     #
#     # def __save(self):
#     #     write_df(df=self.results_df, file_type=self.file_type, save_path=self.save_path)
#     #     self.saved_files.append(self.save_path)
#     #     print(
#     #         f"BENTO annotations appended to video {self.video_name} and saved in {self.save_path}"
#     #     )
#     #
#
# test = BentoAppender(config_path=r"C:\troubleshooting\bento_test\project_folder\project_config.ini",
#                      data_dir=r"C:\troubleshooting\bento_test\bento_files")
# test.run()



__author__ = "Simon Nilsson"

import os
from copy import deepcopy
from typing import Any, Dict, Optional, Union

import pandas as pd

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from simba.mixins.config_reader import ConfigReader
from simba.utils.checks import (check_file_exist_and_readable, check_instance,
                                check_int, check_str, check_that_column_exist,
                                check_valid_boolean)
from simba.utils.data import df_smoother, savgol_smoother
from simba.utils.enums import TagNames
from simba.utils.errors import (DataHeaderError, InvalidInputError,
                                NoFilesFoundError)
from simba.utils.printing import SimbaTimer, log_event, stdout_success
from simba.utils.read_write import (copy_files_to_directory,
                                    find_files_of_filetypes_in_directory,
                                    find_video_of_file, get_fn_ext,
                                    get_video_meta_data, read_df, write_df)

BODY_PART_TYPE = 'body-part'
ANIMAL_TYPE = 'animal'
GAUSSIAN = 'gaussian'
SAVITZKY_GOLAY = 'savitzky_golay'
TIME_WINDOW = 'time_window'

class AdvancedSmoother(ConfigReader):
    """
    Smoothing method that allows different smoothing parameters for different animals or body-parts.
    For example, smooth some body-parts of animals using Savitzky-Golay smoothing, and other body-parts of animals using Gaussian smoothing.

    .. image:: _static/img/AdvancedSmoother.webp
       :width: 600
       :align: center

    .. video:: _static/img/smoothing_example_2.webm
       :width: 800
       :autoplay:
       :loop:

    :parameter str data_dir: path to pose-estimation data in CSV or parquet format
    :parameter str config_path: path to SimBA project config file in Configparser format.
    :parameter Literal type: Level of smoothing: animal or body-part.
    :parameter Dict settings: Smoothing rules for each animal or each animal body-part.
    :parameter bool initial_import_multi_index: If True, the incoming data is multi-index columns dataframes. Use of input data is the ``project_folder/csv/input_csv`` directory. Default: False.
    :parameter bool overwrite: If True, overwrites the input data. If False, then saves a copy input data in datetime-stamped sub-directory.
    :parameter Optional[verbose] bool: If True, prints the progress. Default: True.

    :examples:
    >>> smoother = AdvancedSmoother(data_dir='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/csv/input_csv',
    >>>                             config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
    >>>                             type='animal',
    >>>                             settings={'Simon': {'method': 'Savitzky Golay', 'time_window': 200},
    >>>                                       'JJ': {'method': 'Savitzky Golay', 'time_window': 200}},
    >>>                             initial_import_multi_index=True,
    >>>                             overwrite=False)
    >>> smoother.run()
    >>> SMOOTHING_SETTINGS = {'Simon': {'Ear_left_1': {'method': 'savitzky_golay', 'time_window': 3500},
    >>>                            'Ear_right_1': {'method': 'gaussian', 'time_window': 500},
    >>>                            'Nose_1': {'method': 'savitzky_golay', 'time_window': 2000},
    >>>                            'Lat_left_1': {'method': 'savitzky_golay', 'time_window': 2000},
    >>>                            'Lat_right_1': {'method': 'gaussian', 'time_window': 2000},
    >>>                            'Center_1': {'method': 'savitzky_golay', 'time_window': 2000},
    >>>                            'Tail_base_1': {'method': 'gaussian', 'time_window': 500}},
    >>>                     'JJ': {'Ear_left_2': {'method': 'savitzky_golay', 'time_window': 2000},
    >>>                            'Ear_right_2': {'method': 'savitzky_golay', 'time_window': 500},
    >>>                            'Nose_2': {'method': 'gaussian', 'time_window': 3500},
    >>>                            'Lat_left_2': {'method': 'savitzky_golay', 'time_window': 500},
    >>>                            'Lat_right_2': {'method': 'gaussian', 'time_window': 3500},
    >>>                            'Center_2': {'method': 'gaussian', 'time_window': 2000},
    >>>                            'Tail_base_2': {'method': 'savitzky_golay', 'time_window': 3500}}}
    >>> advanced_smoother = AdvancedSmoother(config_path='/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
    >>>                  data_path='/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/new_data',
    >>>                  settings=SMOOTHING_SETTINGS, type='body-part', multi_index_data=True, overwrite=False)
    >>> advanced_smoother.run()
    """

    def __init__(self,
                 data_path: Union[str, os.PathLike],
                 config_path: Union[str, os.PathLike],
                 settings: Dict[str, Any],
                 type: Optional[Literal["animal", "body-part"]] = 'body-part',
                 verbose: Optional[bool] = True,
                 multi_index_data: Optional[bool] = False,
                 overwrite: Optional[bool] = True):

        ConfigReader.__init__(self, config_path=config_path, read_video_info=False)
        log_event(logger_name=str(self.__class__.__name__), log_type=TagNames.CLASS_INIT.value, msg=f"data_dir: {data_path}, type: {type}, settings: {settings}, initial_import_multi_index: {multi_index_data}, overwrite: {overwrite}",)
        ConfigReader.__init__(self, config_path=config_path, read_video_info=False)
        check_str(name=f'{self.__class__.__name__} type', value=type, options=[ANIMAL_TYPE, BODY_PART_TYPE], raise_error=True)
        if os.path.isfile(data_path):
            check_file_exist_and_readable(file_path=data_path)
            self.file_paths = [data_path]
            self.input_dir = os.path.dirname(data_path)
            self.cpy_dir = os.path.join(os.path.dirname(data_path), f"Pre_Advanced_Interpolation_{self.datetime}")
        elif os.path.isdir(data_path):
            self.file_paths = find_files_of_filetypes_in_directory(directory=data_path, extensions=[f".{self.file_type}"], raise_warning=False,raise_error=True)
            self.cpy_dir = os.path.join(data_path, f"Pre_Advanced_Interpolation_{self.datetime}")
            self.input_dir = data_path
        else:
            raise InvalidInputError(msg=f'{data_path} is not a valid file path or file directory', source=self.__class__.__name__)
        check_valid_boolean(value=[multi_index_data, overwrite], source=self.__class__.__name__, raise_error=True)
        check_instance(source=self.__class__.__name__, instance=settings, accepted_types=(dict,))
        for animal, animal_data in settings.items():
            if type == BODY_PART_TYPE:
                check_instance(source=self.__class__.__name__, instance=animal_data, accepted_types=(dict,))
                for bp_name, bp_data in animal_data.items():
                    check_str(name='body_part', value=bp_name, options=self.project_bps)
                    check_str(name='method', value=bp_data['method'], options=[SAVITZKY_GOLAY, GAUSSIAN])
                    check_int(name='method', value=bp_data[TIME_WINDOW], min_value=1)
            else:
                check_str(name='method', value=animal_data['method'], options=[SAVITZKY_GOLAY, GAUSSIAN])
                check_int(name='method', value=animal_data[TIME_WINDOW], min_value=1)
        self.settings, self.type, self.multi_index_data, self.verbose = settings, type, multi_index_data, verbose
        if type == ANIMAL_TYPE:
            self.__transpose_settings()

        self.overwrite = overwrite
        if not overwrite and not os.path.isdir(self.cpy_dir): os.makedirs(self.cpy_dir)

    def __transpose_settings(self):
        """Helper to transpose settings dict if interpolating per animal, so the same method can be used for both animal and body-part interpolation"""
        transposed_settings = {}
        for animal_name, body_part_data in self.animal_bp_dict.items():
            transposed_settings[animal_name] = {}
            for animal_body_part in body_part_data["X_bps"]:
                transposed_settings[animal_name][animal_body_part[:-2]] = self.settings[animal_name]
        self.settings = transposed_settings

    def __insert_multi_index(self, df: pd.DataFrame) -> pd.DataFrame:
        multi_idx_header = []
        for i in range(len(df.columns)):
            multi_idx_header.append(("IMPORTED_POSE", "IMPORTED_POSE", list(df.columns)[i]))
        df.columns = pd.MultiIndex.from_tuples(multi_idx_header)
        return df

    def run(self):
        for file_cnt, file_path in enumerate(self.file_paths):
            video_timer = SimbaTimer(start=True)
            _, video_name, _ = get_fn_ext(filepath=file_path)
            df = read_df(file_path=file_path, file_type=self.file_type, check_multiindex=self.multi_index_data).fillna(0).reset_index(drop=True)
            if self.verbose: print(f"Smoothing data in video {video_name} ({file_cnt+1}/{len(self.file_paths)})...")
            if len(df.columns) != len(self.bp_col_names):
                raise DataHeaderError(msg=f"The SimBA project suggest the data should have {len(self.bp_col_names)} columns, but the input data has {len(df.columns)} columns", source=self.__class__.__name__)
            df.columns = self.bp_headers
            df[df < 0] = 0
            video_path = find_video_of_file(video_dir=self.video_dir, filename=video_name, warning=False, raise_error=False)
            if video_path is None:
                try:
                    video_meta_data = {}
                    self.video_info_df = self.read_video_info_csv(file_path=self.video_info_path)
                    _, _, fps = self.read_video_info(video_name=video_name)
                    video_meta_data["fps"] = fps
                except:
                    raise NoFilesFoundError(msg=f"No video for file {video_name} found in SimBA project. Import the video before doing smoothing. To perform smoothing, SimBA needs the video fps from the video itself OR the logs/video_info.csv file in order to read the video FPS.", source=self.__class__.__name__)
            else:
                video_meta_data = get_video_meta_data(video_path=video_path)
            out_df = deepcopy(df)
            for animal_name, animal_body_parts in self.settings.items():
                for bp, smoothing_setting in animal_body_parts.items():
                    if self.verbose: print(f"Smoothing body-part {bp} in video {video_name} using method {smoothing_setting['method']} (time window: {smoothing_setting[TIME_WINDOW]}ms)...")
                    check_that_column_exist(df=df, column_name=[f"{bp}_x", f"{bp}_y"], file_name=file_path)
                    bp_df = df[[f"{bp}_x", f"{bp}_y"]]
                    if smoothing_setting['method'] == SAVITZKY_GOLAY:
                        bp_df = savgol_smoother(data=bp_df, fps=video_meta_data['fps'], time_window=smoothing_setting[TIME_WINDOW], source=f'{file_path} {bp}')
                    else:
                        bp_df = df_smoother(data=bp_df, fps=video_meta_data['fps'], time_window=smoothing_setting[TIME_WINDOW], source=f'{file_path} {bp}')
                    out_df[[f"{bp}_x", f"{bp}_y"]] = bp_df
            if self.multi_index_data:
                out_df = self.__insert_multi_index(df=out_df)
            if not self.overwrite:
                copy_files_to_directory(file_paths=[file_path], dir=self.cpy_dir, verbose=False)
            write_df(df=out_df, file_type=self.file_type, save_path=file_path, multi_idx_header=self.multi_index_data)
            video_timer.stop_timer()
            print(f'Smoothing video {video_name} complete ({file_cnt+1}/{len(self.file_paths)}). Elapsed time {video_timer.elapsed_time_str}s')
        self.timer.stop_timer()
        if self.overwrite:
            msg = f"Advanced smoothing complete. Data saved in {self.input_dir}."
        else:
            msg = f"Advanced smoothing complete. Data saved in {self.input_dir}. Original data saved in {self.cpy_dir}."
        stdout_success(msg=msg, elapsed_time=self.timer.elapsed_time_str, source=self.__class__.__name__)

# SMOOTHING_SETTINGS = {'Simon': {'Ear_left_1': {'method': 'savitzky_golay', 'time_window': 3500},
#                                'Ear_right_1': {'method': 'gaussian', 'time_window': 500},
#                                'Nose_1': {'method': 'savitzky_golay', 'time_window': 2000},
#                                'Lat_left_1': {'method': 'savitzky_golay', 'time_window': 2000},
#                                'Lat_right_1': {'method': 'gaussian', 'time_window': 2000},
#                                'Center_1': {'method': 'savitzky_golay', 'time_window': 2000},
#                                'Tail_base_1': {'method': 'gaussian', 'time_window': 500}},
#                         'JJ': {'Ear_left_2': {'method': 'savitzky_golay', 'time_window': 2000},
#                                'Ear_right_2': {'method': 'savitzky_golay', 'time_window': 500},
#                                'Nose_2': {'method': 'gaussian', 'time_window': 3500},
#                                'Lat_left_2': {'method': 'savitzky_golay', 'time_window': 500},
#                                'Lat_right_2': {'method': 'gaussian', 'time_window': 3500},
#                                'Center_2': {'method': 'gaussian', 'time_window': 2000},
#                                'Tail_base_2': {'method': 'savitzky_golay', 'time_window': 3500}}}


#SMOOTHING_SETTINGS = {'Animal_1': {'method': 'savitzky_golay', 'time_window': 3500}, 'Animal_2': {'method': 'savitzky_golay', 'time_window': 3500}}
# advanced_smoother = AdvancedSmoother(config_path='/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                      data_path='/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/new_data',
#                      settings=SMOOTHING_SETTINGS, type='body-part', multi_index_data=True, overwrite=False)
#
# advanced_smoother.run()



import pandas as pd

pd.options.mode.chained_assignment = None
import os
from copy import deepcopy
from typing import List, Optional, Union

import numpy as np

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from simba.mixins.config_reader import ConfigReader
from simba.utils.checks import (check_file_exist_and_readable, check_str,
                                check_valid_lst)
from simba.utils.data import animal_interpolator, body_part_interpolator
from simba.utils.enums import TagNames
from simba.utils.errors import DataHeaderError, InvalidInputError
from simba.utils.printing import SimbaTimer, log_event, stdout_success
from simba.utils.read_write import (copy_files_to_directory,
                                    find_files_of_filetypes_in_directory,
                                    get_fn_ext, read_df, write_df)


class Interpolate(ConfigReader):
    """
    Interpolate missing body-parts in pose-estimation data. "Missing" is defined as either (i) when a single body-parts is None, or
    when all body-parts belonging to an animal are identical (i.e., the same 2D coordinate or all None).

    .. image:: _static/img/interpolation_comparison.png
       :width: 500
       :align: center

    .. note::
       `Interpolation tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/Scenario1.md#to-import-multiple-dlc-csv-files>`__.

    .. important::
        The interpolated data overwrites the original data on disk. If the original data is required, pass ``copy_originals = True`` to save a copy of the original data.


    :param Union[str, os.PathLike] config_path: path to SimBA project config file in Configparser format.
    :param Union[str, os.PathLike] data_path: Path to a directory, path to a file, or a list of file paths to files with pose-estimation data in CSV or parquet format.
    :param Optional[Literal['body-parts', 'animals']] type: If 'animals', then interpolation is performed when all body-parts belonging to an animal are identical (i.e., the same 2D coordinate or all None). If 'body-parts` then all body-parts that are None will be interpolated. Default: body-parts.
    :param Optional[Literal['nearest', 'linear', 'quadratic']] method: If 'animals', then interpolation is performed when all body-parts belonging to an animal are identical (i.e., the same 2D coordinate or all None). If 'body-parts` then all body-parts that are None will be interpolated. Default: body-parts.
    :param Optional[bool] multi_index_df_headers: If truth-like, then the input data is anticipated to have multiple header columns, and output columns will have multiple header columns. Default: False.
    :param Optional[bool] copy_originals: If truth-like, then the pre-interpolated, original data, will be bo stored in a subdirectory of the original data. The subdirectory is named according to the type of interpolation and datetime of the operation.

    :example:
    >>> interpolator = Interpolate(config_path='/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini', data_path='/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/csv/input_csv/test', type='body-parts', multi_index_df_headers=True, copy_originals=True)
    >>> interpolator.run()

    """
    def __init__(self,
                 config_path: Union[str, os.PathLike],
                 data_path: Union[str, os.PathLike, List[Union[str, os.PathLike]]],
                 type: Optional[Literal['body-parts', 'animals']] = 'body-parts',
                 method: Optional[Literal['nearest', 'linear', 'quadratic']] = 'nearest',
                 multi_index_df_headers: Optional[bool] = False,
                 copy_originals: Optional[bool] = False) -> None:

        log_event(logger_name=str(self.__class__.__name__), log_type=TagNames.CLASS_INIT.value, msg=self.create_log_msg_from_init_args(locals=locals()))
        ConfigReader.__init__(self, config_path=config_path, read_video_info=False)
        check_str(name=f'{self.__class__.__name__} type', value=type.lower(), options=('body-parts', 'animals'))
        check_str(name=f'{self.__class__.__name__} method', value=method.lower(), options=('nearest', 'linear', 'quadratic'))
        if isinstance(data_path, list):
            check_valid_lst(data=data_path, source=self.__class__.__name__, valid_dtypes=(str,))
            for i in data_path: check_file_exist_and_readable(file_path=i)
            self.file_paths = deepcopy(data_path)
        elif os.path.isdir(data_path):
            self.file_paths = find_files_of_filetypes_in_directory(directory=data_path, extensions=[f'.{self.file_type}'], raise_error=True)
        elif os.path.isfile(data_path):
            check_file_exist_and_readable(file_path=data_path)
            self.file_paths = [data_path]
        else:
            raise InvalidInputError(msg=f'{data_path} is not a valid data directory, or a valid file path, or a valid list of file paths', source=self.__class__.__name__)
        if copy_originals:
            self.originals_dir = os.path.join(os.path.dirname(self.file_paths[0]), f"Pre_{method}_{type}_interpolation_{self.datetime}")
            os.makedirs(self.originals_dir)
        self.type, self.method, self.multi_index_df_headers, self.copy_originals = type.lower(), method.lower(), multi_index_df_headers, copy_originals

    def __insert_multiindex_header(self, df: pd.DataFrame):
        multi_idx_header = []
        for i in range(len(df.columns)):
            multi_idx_header.append(("IMPORTED_POSE", "IMPORTED_POSE", list(df.columns)[i]))
        df.columns = pd.MultiIndex.from_tuples(multi_idx_header)
        return df

    def run(self):
        print(f'Running interpolation on {len(self.file_paths)} data files...')
        for file_cnt, file_path in enumerate(self.file_paths):
            video_timer = SimbaTimer(start=True)
            _, self.video_name, _ = get_fn_ext(filepath=file_path)
            df = read_df(file_path=file_path, file_type=self.file_type, check_multiindex=self.multi_index_df_headers)
            if self.multi_index_df_headers:
                if len(df.columns) != len(self.bp_headers):
                    raise DataHeaderError( msg=f"The file {file_path} contains {len(df.columns)} columns, but your SimBA project expects {len(self.bp_headers)} columns representing {int(len(self.bp_headers) / 3)} body-parts (x, y, p). Check that the {self.body_parts_path} lists the correct body-parts associated with the project", source=self.__class__.__name__)
                df.columns = self.bp_headers
            df = df.apply(pd.to_numeric, errors="coerce").fillna(0.0)
            df[df < 0] = 0.0
            if self.type == 'animals':
                df = animal_interpolator(df=df, animal_bp_dict=self.animal_bp_dict, source=file_path, method=self.method)
            else:
                df = body_part_interpolator(df=df, animal_bp_dict=self.animal_bp_dict, source=file_path, method=self.method)
            if self.multi_index_df_headers:
                df = self.__insert_multiindex_header(df=df)
            if self.copy_originals:
                copy_files_to_directory(file_paths=[file_path], dir=self.originals_dir)
            write_df(df=df.astype(np.float32), file_type=self.file_type, save_path=file_path, multi_idx_header=self.multi_index_df_headers)
            video_timer.stop_timer()
            print(f"Video {self.video_name} interpolated (elapsed time {video_timer.elapsed_time_str}) ...")
        self.timer.stop_timer()
        if self.copy_originals:
            msg = f"{len(self.file_paths)} data file(s) interpolated using {self.type} {self.method} methods. Originals saved in {self.originals_dir} directory."
        else:
            msg = f"{len(self.file_paths)} data file(s) interpolated using {self.type} {self.method} methods."
        stdout_success(msg=msg, elapsed_time=self.timer.elapsed_time_str, source=self.__class__.__name__)


__author__ = "Simon Nilsson"

import os
from tkinter import *
from typing import Union

from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.ui.tkinter_functions import FileSelect
from simba.unsupervised.cluster_frequentist_calculator import \
    ClusterFrequentistCalculator
from simba.utils.checks import check_file_exist_and_readable
from simba.utils.enums import Formats


class ClusterFrequentistStatisticsPopUp(PopUpMixin, ConfigReader):
    def __init__(self, config_path: Union[str, os.PathLike]):
        PopUpMixin.__init__(self, title="CLUSTER FREQUENTIST STATISTICS")
        ConfigReader.__init__(self, config_path=config_path)
        self.descriptive_stats_var = BooleanVar(value=True)
        self.kruskal_wallis_var = BooleanVar(value=True)
        self.oneway_anova_var = BooleanVar(value=True)
        self.tukey_var = BooleanVar(value=True)
        self.use_scaled_var = BooleanVar(value=False)

        self.data_frm = LabelFrame(
            self.main_frm,
            text="DATA",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
        )
        self.model_select = FileSelect(
            self.data_frm,
            "CLUSTERER PATH:",
            lblwidth=25,
            initialdir=self.project_path,
            file_types=[("Cluster model pickle", f"*.{Formats.PICKLE.value}")],
        )
        self.data_frm.grid(row=0, column=0, sticky=NW)
        self.model_select.grid(row=0, column=0, sticky=NW)

        self.stats_frm = LabelFrame(
            self.main_frm,
            text="STATISTICS",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
        )
        self.descriptive_stats_cb = Checkbutton(
            self.stats_frm,
            text="CLUSTER DESCRIPTIVE STATISTICS",
            variable=self.descriptive_stats_var,
        )
        self.oneway_anova_cb = Checkbutton(
            self.stats_frm,
            text="CLUSTER FEATURE ONE-WAY ANOVA",
            variable=self.oneway_anova_var,
        )
        self.feature_tukey_posthoc_cb = Checkbutton(
            self.stats_frm,
            text="CLUSTER FEATURE POST-HOC (TUKEY)",
            variable=self.tukey_var,
        )
        self.use_scaled_cb = Checkbutton(
            self.stats_frm,
            text="USE SCALED FEATURE VALUES",
            variable=self.use_scaled_var,
        )

        self.kruskal_wallis_cb = Checkbutton(
            self.stats_frm,
            text="KRUSKAL-WALLIS",
            variable=self.kruskal_wallis_var,
        )

        self.stats_frm.grid(row=1, column=0, sticky=NW)
        self.descriptive_stats_cb.grid(row=0, column=0, sticky=NW)
        self.oneway_anova_cb.grid(row=1, column=0, sticky=NW)
        self.kruskal_wallis_cb.grid(row=2, column=0, sticky=NW)
        self.feature_tukey_posthoc_cb.grid(row=3, column=0, sticky=NW)
        self.use_scaled_cb.grid(row=4, column=0, sticky=NW)
        self.create_run_frm(run_function=self.run)
        self.main_frm.mainloop()

    def run(self):
        check_file_exist_and_readable(self.model_select.file_path)
        settings = {
            "scaled": self.use_scaled_var.get(),
            "anova": self.oneway_anova_var.get(),
            "tukey_posthoc": self.tukey_var.get(),
            "kruskal_wallis": self.kruskal_wallis_var.get(),
            "descriptive_statistics": self.descriptive_stats_var.get(),
        }
        calculator = ClusterFrequentistCalculator(
            config_path=self.config_path,
            data_path=self.model_select.file_path,
            settings=settings,
        )
        calculator.run()


# _ = ClusterFrequentistStatisticsPopUp(config_path='/Users/simon/Desktop/envs/simba/troubleshooting/NG_Unsupervised/project_folder/project_config.ini')



__author__ = "Simon Nilsson"

import warnings

warnings.filterwarnings("ignore", category=FutureWarning)
warnings.filterwarnings("ignore", category=DeprecationWarning)
import os
import warnings
from typing import Optional, Union

import cv2
import matplotlib.pyplot as plt
import numpy as np

from simba.mixins.config_reader import ConfigReader
from simba.mixins.plotting_mixin import PlottingMixin
from simba.mixins.train_model_mixin import TrainModelMixin
from simba.utils.checks import (check_file_exist_and_readable, check_float,
                                check_if_keys_exist_in_dict, check_int,
                                check_valid_boolean,
                                check_video_and_data_frm_count_align)
from simba.utils.data import plug_holes_shortest_bout
from simba.utils.enums import Formats, TagNames, TextOptions
from simba.utils.printing import log_event, stdout_success
from simba.utils.read_write import (get_fn_ext, get_video_meta_data, read_df,
                                    read_pickle, write_df)

plt.interactive(True)
plt.ioff()
warnings.simplefilter(action="ignore", category=FutureWarning)


class ValidateModelOneVideo(ConfigReader, PlottingMixin, TrainModelMixin):
    """
    Create classifier validation video for a single input video. Results are stored in the
    `project_folder/frames/output/validation directory`.

    .. note::
       For improved run-time, see :meth:`simba.sing_run_model_validation_video_mp.ValidateModelOneVideoMultiprocess` for multiprocess class.

    :param str config_path: path to SimBA project config file in Configparser format
    :param str feature_file_path: path to SimBA file (parquet or CSV) containing pose-estimation and feature fields.
    :param str model_path: path to pickled classifier object
    :param float discrimination_threshold: classification threshold.
    :param int shortest_bout: Allowed classified bout length expressed in milliseconds. E.g., `1000` will shift frames classified
        as containing the behavior, but occuring in a bout shorter than `1000`, from `target present to `target absent`.
    :param str create_gantt:
        If SimBA should create gantt charts alongside the validation video. OPTIONS: 'None', 'Gantt chart: final frame only (slightly faster)',
        'Gantt chart: video'.
    :param dict settings: User style settings for video. E.g., {'pose': True, 'animal_names': True, 'styles': None}

    :example:
    >>> test = ValidateModelOneVideo(config_path=r'/Users/simon/Desktop/envs/troubleshooting/naresh/project_folder/project_config.ini',
    >>>                                 feature_file_path=r'/Users/simon/Desktop/envs/troubleshooting/naresh/project_folder/csv/features_extracted/SF2.csv',
    >>>                                 model_path='/Users/simon/Desktop/envs/troubleshooting/naresh/models/generated_models/Top.sav',
    >>>                                 discrimination_threshold=0.6,
    >>>                                 shortest_bout=50,
    >>>                                 settings={'pose': True, 'animal_names': True, 'styles': None},
    >>>                                 create_gantt='Gantt chart: final frame only (slightly faster)')
    >>> test.run()
    """

    def __init__(self,
                 config_path: Union[str, os.PathLike],
                 feature_path: Union[str, os.PathLike],
                 model_path: Union[str, os.PathLike],
                 show_pose: bool = True,
                 show_animal_names: bool = False,
                 font_size: Optional[bool] = None,
                 circle_size: Optional[int] = None,
                 text_spacing: Optional[int] = None,
                 text_thickness: Optional[int] = None,
                 text_opacity: Optional[float] = None,
                 discrimination_threshold: Optional[float] = 0.0,
                 shortest_bout: Optional[int] = 0.0,
                 create_gantt: Optional[Union[None, int]] = None):

        ConfigReader.__init__(self, config_path=config_path)
        PlottingMixin.__init__(self)
        TrainModelMixin.__init__(self)
        log_event(logger_name=str(__class__.__name__), log_type=TagNames.CLASS_INIT.value, msg=self.create_log_msg_from_init_args(locals=locals()))
        check_file_exist_and_readable(file_path=config_path)
        check_file_exist_and_readable(file_path=feature_path)
        check_file_exist_and_readable(file_path=model_path)
        check_valid_boolean(value=[show_pose], source=f'{self.__class__.__name__} show_pose', raise_error=True)
        check_valid_boolean(value=[show_animal_names], source=f'{self.__class__.__name__} show_animal_names', raise_error=True)
        if font_size is not None: check_int(name=f'{self.__class__.__name__} font_size', value=font_size)
        if circle_size is not None: check_int(name=f'{self.__class__.__name__} circle_size', value=circle_size)
        if text_spacing is not None: check_int(name=f'{self.__class__.__name__} text_spacing', value=text_spacing)
        if text_opacity is not None: check_float(name=f'{self.__class__.__name__} text_opacity', value=text_opacity, min_value=0.1)
        if text_thickness is not None: check_float(name=f'{self.__class__.__name__} text_thickness', value=text_thickness, min_value=0.1)
        check_float(name=f"{self.__class__.__name__} discrimination_threshold", value=discrimination_threshold, min_value=0, max_value=1.0)
        check_int(name=f"{self.__class__.__name__} shortest_bout", value=shortest_bout, min_value=0)
        if create_gantt is not None:
            check_int(name=f"{self.__class__.__name__} create gantt", value=create_gantt, max_value=2, min_value=1)
        if not os.path.exists(self.single_validation_video_save_dir):
            os.makedirs(self.single_validation_video_save_dir)
        _, self.feature_filename, ext = get_fn_ext(feature_path)
        self.video_path = self.find_video_of_file(self.video_dir, self.feature_filename)
        self.video_meta_data = get_video_meta_data(video_path=self.video_path)
        self.clf_name, self.feature_file_path = (os.path.basename(model_path).replace(".sav", ""), feature_path)
        self.vid_output_path = os.path.join(self.single_validation_video_save_dir, f"{self.feature_filename} {self.clf_name}.mp4")
        self.clf_data_save_path = os.path.join(self.clf_data_validation_dir, f"{self.feature_filename }.csv")
        self.show_pose, self.show_animal_names = show_pose, show_animal_names
        self.font_size, self.circle_size, self.text_spacing = font_size, circle_size, text_spacing
        self.text_opacity, self.text_thickness = text_opacity, text_thickness
        self.clf = read_pickle(data_path=model_path, verbose=True)
        self.data_df = read_df(feature_path, self.file_type)
        self.feature_path =  feature_path
        self.x_df = self.drop_bp_cords(df=self.data_df)
        self.discrimination_threshold, self.shortest_bout, self.create_gantt = float(discrimination_threshold), shortest_bout, create_gantt
        check_video_and_data_frm_count_align(video=self.video_path, data=self.data_df, name=self.feature_filename, raise_error=False)

    def _get_styles(self):
        self.video_text_thickness = TextOptions.TEXT_THICKNESS.value if self.text_thickness is None else int(max(self.text_thickness, 1))
        longest_str = str(max(['TIMERS:', 'ENSEMBLE PREDICTION:'] + self.clf_names, key=len))
        optimal_font_size, _, optimal_spacing_scale = self.get_optimal_font_scales(text=longest_str, accepted_px_width=int(self.video_meta_data["width"] / 3), accepted_px_height=int(self.video_meta_data["height"] / 10), text_thickness=self.video_text_thickness)
        optimal_circle_size = self.get_optimal_circle_size(frame_size=(self.video_meta_data["width"], self.video_meta_data["height"]), circle_frame_ratio=100)
        self.video_circle_size = optimal_circle_size if self.circle_size is None else int(self.circle_size)
        self.video_font_size = optimal_font_size if self.font_size is None else self.font_size
        self.video_space_size = optimal_spacing_scale if self.text_spacing is None else int(max(self.text_spacing, 1))
        self.video_text_opacity = 0.8 if self.text_opacity is None else float(self.text_opacity)

    def run(self):
        self.prob_col_name = f"Probability_{self.clf_name}"
        self.data_df[self.prob_col_name] = self.clf_predict_proba(clf=self.clf, x_df=self.x_df, model_name=self.clf_name, data_path=self.feature_path)
        self.data_df[self.clf_name] = np.where(self.data_df[self.prob_col_name] > self.discrimination_threshold, 1, 0)
        if self.shortest_bout > 1:
            self.data_df = plug_holes_shortest_bout(data_df=self.data_df, clf_name=self.clf_name, fps=self.video_meta_data['fps'], shortest_bout=self.shortest_bout)
        _ = write_df(df=self.data_df, file_type=self.file_type, save_path=self.clf_data_save_path)
        print(f"Behavior predictions created for model {self.clf_name} and video {self.feature_filename}...")
        cap = cv2.VideoCapture(self.video_path)
        fourcc, font = cv2.VideoWriter_fourcc(*"mp4v"), cv2.FONT_HERSHEY_COMPLEX
        if self.create_gantt is not None:
            self.bouts_df = self.get_bouts_for_gantt(data_df=self.data_df, clf_name=self.clf_name, fps=self.video_meta_data['fps'])
            self.final_gantt_img = self.create_gantt_img(bouts_df=self.bouts_df, clf_name=self.clf_name, image_index=len(self.data_df), fps=self.video_meta_data['fps'], gantt_img_title=f"Behavior gantt chart (entire session, length (s): {self.video_meta_data['video_length_s']}, frames: {self.video_meta_data['frame_count']})")
            self.final_gantt_img = self.resize_gantt(self.final_gantt_img, self.video_meta_data["height"])
            video_size = (int(self.video_meta_data["width"] + self.final_gantt_img.shape[1]), int(self.video_meta_data["height"]))
            writer = cv2.VideoWriter(self.vid_output_path, fourcc, self.video_meta_data["fps"], video_size)
        else:
            video_size = (int(self.video_meta_data["width"]), int(self.video_meta_data["height"]))
            writer = cv2.VideoWriter(self.vid_output_path, fourcc, self.video_meta_data["fps"], video_size)

        self._get_styles()
        self.data_df = self.data_df.head(min(len(self.data_df), self.video_meta_data["frame_count"]))
        frm_cnt, clf_frm_cnt = 0, 0
        print(f"Creating single core validation visualization for video {self.video_meta_data['video_name']} and classifier {self.clf_name}...")
        while (cap.isOpened()) and (frm_cnt < len(self.data_df)):
            ret, frame = cap.read()
            clf_val = int(self.data_df.loc[frm_cnt, self.clf_name])
            clf_frm_cnt += clf_val
            if self.show_pose:
                for animal_cnt, (animal_name, animal_data) in enumerate(self.animal_bp_dict.items()):
                    for bp_cnt, bp in enumerate(range(len(animal_data["X_bps"]))):
                        x_header, y_header = (animal_data["X_bps"][bp], animal_data["Y_bps"][bp])
                        animal_cords = tuple(self.data_df.loc[self.data_df.index[frm_cnt], [x_header, y_header]])
                        cv2.circle(frame, (int(animal_cords[0]), int(animal_cords[1])), self.video_circle_size, self.clr_lst[animal_cnt][bp_cnt], -1)
            if self.show_animal_names:
                for animal_cnt, (animal_name, animal_data) in enumerate(self.animal_bp_dict.items()):
                    x_header, y_header = ( animal_data["X_bps"][0], animal_data["Y_bps"][0])
                    animal_cords = tuple(self.data_df.loc[self.data_df.index[frm_cnt], [x_header, y_header]])
                    cv2.putText(frame, animal_name, (int(animal_cords[0]), int(animal_cords[1])), self.font, self.video_font_size, self.clr_lst[animal_cnt][0], self.video_text_thickness)
            target_timer = round((1 / self.video_meta_data['fps']) * clf_frm_cnt, 2)
            frame = PlottingMixin().put_text(img=frame, text="TIMER:", pos=(TextOptions.BORDER_BUFFER_Y.value, self.video_space_size), font_size=self.video_font_size, font_thickness=TextOptions.TEXT_THICKNESS.value, text_color=(255, 255, 255))
            addSpacer = 2
            frame = PlottingMixin().put_text(img=frame, text=(f"{self.clf_name} {target_timer}s"), pos=(TextOptions.BORDER_BUFFER_Y.value, self.video_space_size * addSpacer), font_size=self.video_font_size, font_thickness=TextOptions.TEXT_THICKNESS.value, text_color=(255, 255, 255))
            addSpacer += 1
            frame = PlottingMixin().put_text(img=frame, text="ENSEMBLE PREDICTION:", pos=(TextOptions.BORDER_BUFFER_Y.value, self.video_space_size * addSpacer), font_size=self.video_font_size, font_thickness=TextOptions.TEXT_THICKNESS.value, text_color=(255, 255, 255))
            addSpacer += 2
            if clf_val == 1:
                frame = PlottingMixin().put_text(img=frame, text=self.clf_name, pos=(TextOptions.BORDER_BUFFER_Y.value, self.video_space_size * addSpacer), font_size=self.video_font_size, font_thickness=TextOptions.TEXT_THICKNESS.value, text_color=TextOptions.COLOR.value)
                addSpacer += 1
            if self.create_gantt == 1:
                frame = np.concatenate((frame, self.final_gantt_img), axis=1)
            elif self.create_gantt == 2:
                gantt_img = self.create_gantt_img(bouts_df=self.bouts_df, clf_name=self.clf_name, image_index=len(self.data_df), fps=self.video_meta_data['fps'], gantt_img_title=f"Behavior gantt chart: {self.clf_name}")
                gantt_img = self.resize_gantt(gantt_img, self.video_meta_data["height"])
                frame = np.concatenate((frame, gantt_img), axis=1)
            frame = cv2.resize(frame, video_size, interpolation=cv2.INTER_LINEAR)
            writer.write(np.uint8(frame))
            print(f"Frame created: for video {self.feature_filename} ({frm_cnt + 1} / {len(self.data_df)})...")
            frm_cnt += 1

        cap.release()
        writer.release()
        self.timer.stop_timer()
        stdout_success(msg=f"Validation video saved at {self.vid_output_path}", elapsed_time=self.timer.elapsed_time_str, source=self.__class__.__name__)


# test = ValidateModelOneVideo(config_path=r"D:\troubleshooting\mitra\project_folder\project_config.ini",
#                              feature_path=r"D:\troubleshooting\mitra\project_folder\csv\features_extracted\592_MA147_CNO1_0515.csv",
#                              model_path=r"C:\troubleshooting\mitra\models\generated_models\grroming_undersample_2_1000\grooming.sav",
#                              create_gantt=1,
#                              show_pose=True,
#                              show_animal_names=True)
# test.run()
# ValidationVideoPopUp(config_path=r"C:\troubleshooting\mitra\project_folder\project_config.ini",
#                      feature_path=r"C:\troubleshooting\mitra\project_folder\csv\features_extracted\501_MA142_Gi_CNO_0521.csv",
#                      model_path=r"C:\troubleshooting\mitra\models\generated_models\grroming_undersample_2_1000\grooming.sav",
#                      discrimination_threshold=0.4,
#                      shortest_bout=500)







# test = ValidateModelOneVideo(config_path=r'/Users/simon/Desktop/envs/simba/troubleshooting/mouse_open_field/project_folder/project_config.ini',
#                              feature_file_path='/Users/simon/Desktop/envs/simba/troubleshooting/mouse_open_field/project_folder/csv/features_extracted/SI_DAY3_308_CD1_PRESENT.csv',
#                              model_path='/Users/simon/Desktop/envs/simba/troubleshooting/mouse_open_field/models/generated_models/Running.sav',
#                              discrimination_threshold=0.6,
#                              shortest_bout=50,
#                              settings={'pose': True, 'animal_names': True, 'styles': None},
#                              create_gantt=None)
# test.run()

# test = ValidateModelOneVideo(config_path=r'/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                              feature_file_path='/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/csv/features_extracted/Together_1.csv',
#                              model_path='/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/models/generated_models/Attack.sav',
#                              discrimination_threshold=0.6,
#                              shortest_bout=50,
#                              settings={'pose': True, 'animal_names': True, 'styles': None},
#                              create_gantt=2)
# test.run()


# test = ValidateModelOneVideoMultiprocess(config_path=r'/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                              feature_file_path='/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/csv/features_extracted/Together_1.csv',
#                              model_path='/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/models/generated_models/Attack.sav',
#                              discrimination_threshold=0.6,
#                              shortest_bout=50,
#                              cores=6,
#                              settings={'pose': True, 'animal_names': True, 'styles': None},
#                              create_gantt=None)
# test.run()


# test = ValidateModelOneVideo(config_path=r'/Users/simon/Desktop/envs/troubleshooting/dam_nest-c-only_ryan/project_folder/project_config.ini',
#                              feature_file_path='/Users/simon/Desktop/envs/troubleshooting/dam_nest-c-only_ryan/project_folder/csv/features_extracted/LBNF2_Ctrl_P04_4_2021-03-18_19-49-46c.csv',
#                              model_path='/Users/simon/Desktop/envs/troubleshooting/dam_nest-c-only_ryan/models/dam_in_nest.sav',
#                              discrimination_threshold=0.6,
#                              shortest_bout=50,
#                              settings={'pose': True, 'animal_names': True, 'styles': None},
#                              create_gantt=None)
# test.run()


# test = ValidateModelOneVideo(config_path=r'/Users/simon/Desktop/envs/troubleshooting/naresh/project_folder/project_config.ini',
#                              feature_file_path=r'/Users/simon/Desktop/envs/troubleshooting/naresh/project_folder/csv/features_extracted/SF2.csv',
#                              model_path='/Users/simon/Desktop/envs/troubleshooting/naresh/models/generated_models/Top.sav',
#                              discrimination_threshold=0.6,
#                              shortest_bout=50,
#                              settings={'pose': True, 'animal_names': True, 'styles': None},
#                              create_gantt='Gantt chart: final frame only (slightly faster)')
# test.run()


# test = ValidateModelOneVideo(config_path=r'/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                              feature_file_path=r'/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/csv/features_extracted/Together_1.csv',
#                              model_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/models/generated_models/Attack.sav',
#                              discrimination_threshold=0.6,
#                              shortest_bout=50,
#                              settings={'pose': True, 'animal_names': True, 'styles': None},
#                              create_gantt='Gantt chart: final frame only (slightly faster)')
# test.run()


# test.perform_clf()
# test.plug_small_bouts()
# test.save_classification_data()
# test.create_video()

# test = ValidateModelOneVideo(ini_path='/Users/simon/Desktop/troubleshooting/Zebrafish/project_folder/project_config.ini',
#                  feature_file_path=r'/Users/simon/Desktop/troubleshooting/Zebrafish/project_folder/csv/features_extracted/20200730_AB_7dpf_850nm_0002.csv',
#                  model_path='/Users/simon/Desktop/troubleshooting/Zebrafish/models/generated_models/Rheotaxis.sav',
#                  d_threshold=0,
#                  shortest_bout=50,
#                  create_gantt='Gantt chart: final frame only (slightly faster)')
# test.perform_clf()
# test.plug_small_bouts()
# test.save_classification_data()
# test.create_video()


# inifile = r"Z:\DeepLabCut\DLC_extract\Troubleshooting\DLC_two_mice\project_folder\project_config.ini"
# featuresPath = r"Z:\DeepLabCut\DLC_extract\Troubleshooting\DLC_two_mice\project_folder\csv\features_extracted\Together_1.csv"
# modelPath = r"Z:\DeepLabCut\DLC_extract\Troubleshooting\DLC_two_mice\models\generated_models\Attack.sav"
#
# inifile = r"Z:\DeepLabCut\DLC_extract\Troubleshooting\Zebrafish\project_folder\project_config.ini"
# featuresPath = r"Z:\DeepLabCut\DLC_extract\Troubleshooting\Zebrafish\project_folder\csv\features_extracted\20200730_AB_7dpf_850nm_0002.csv"
# modelPath = r"Z:\DeepLabCut\DLC_extract\Troubleshooting\Zebrafish\models\validations\model_files\Rheotaxis_1.sav"
#
# dt = 0.4
# sb = 67
# generategantt = 'Gantt chart: video'
#
# test = ValidateModelOneVideo(ini_path=inifile,feature_file_path=featuresPath,model_path=modelPath,d_threshold=dt,shortest_bout=sb, create_gantt=generategantt)
# test.perform_clf()
# test.plug_small_bouts()
# test.save_classification_data()
# test.create_video()

# cv2.imshow('Window', frame)
# key = cv2.waitKey(3000)
# if key == 27:
#     cv2.destroyAllWindows()
